import logging

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.pyplot import cm
from tqdm.auto import tqdm

import openpnm as op

logger = logging.getLogger(__name__)


__all__ = [
    'plot_connections',
    'plot_coordinates',
    'plot_networkx',
    'plot_tutorial',
    'plot_notebook',
    'plot_vispy',
    'set_mpl_style',
]


def plot_connections(network,
                     throats=None,
                     ax=None,
                     size_by=None,
                     color_by=None,
                     label_by=None,
                     cmap='jet',
                     color='b',
                     alpha=1.0,
                     linestyle='solid',
                     linewidth=1,
                     **kwargs):  # pragma: no cover
    r"""
    Produce a 3D plot of the network topology.

    This shows how throats connect for quick visualization without having
    to export data to veiw in Paraview.

    Parameters
    ----------
    network : Network
        The network whose topological connections to plot
    throats : array_like (optional)
        The list of throats to plot if only a sub-sample is desired.  This is
        useful for inspecting a small region of the network.  If no throats are
        specified then all throats are shown.
    fig : Matplotlib figure handle and line property arguments (optional)
        If a ``fig`` is supplied, then the topology will be overlaid on this
        plot.  This makes it possible to combine coordinates and connections,
        and to color throats differently for instance.
    size_by : array_like (optional)
        An ndarray of throat values (e.g. alg['throat.rate']).  These
        values are used to scale the ``linewidth``, so if the lines are too
        thin, then increase ``linewidth``.
    color_by : array_like (optional)
        An ndarray of throat values (e.g. alg['throat.rate']).
    label_by : array_like (optional)
        An array or list of values to use as labels
    cmap : str or cmap object (optional)
        The matplotlib colormap to use if specfying a throat property
        for ``color_by``
    color : str, optional (optional)
        A matplotlib named color (e.g. 'r' for red).
    alpha : float (optional)
        The transparency of the lines, with 1 being solid and 0 being invisible
    linestyle : str (optional)
        Can be one of {'solid', 'dashed', 'dashdot', 'dotted'}.  Default is
        'solid'.
    linewidth : float (optional)
        Controls the thickness of drawn lines.  Is used to scale the thickness
        if ``size_by`` is given. Default is 1. If a value is provided for
        ``size_by`` then they are used to scale the ``linewidth``.
    font : dict
        A dictionary of key-value pairs that are used to control the font
        appearance if `label_by` is provided.
    **kwargs : dict
        All other keyword arguments are passed on to the ``Line3DCollection``
        class of matplotlib, so check their documentation for additional
        formatting options.

    Returns
    -------
    lc : LineCollection or Line3DCollection
        Matplotlib object containing the lines representing the throats.

    Notes
    -----
    To create a single plot containing both pore coordinates and throats,
    consider creating an empty figure and then pass the ``ax`` object as
    an argument to ``plot_connections`` and ``plot_coordinates``.
    Otherwise, each call to either of these methods creates a new figure.

    See Also
    --------
    plot_coordinates

    Examples
    --------
    >>> import openpnm as op
    >>> import matplotlib as mpl
    >>> import matplotlib.pyplot as plt
    >>> mpl.use('Agg')
    >>> pn = op.network.Cubic(shape=[10, 10, 3])
    >>> pn.add_boundary_pores()
    >>> Ts = pn.throats('*boundary', mode='not')  # find internal throats
    >>> fig, ax = plt.subplots()  # create empty figure
    >>> _ = op.visualization.plot_connections(network=pn,
    ...                                       throats=Ts)  # plot internal throats
    >>> Ts = pn.throats('*boundary')  # find boundary throats
    >>> _ = op.visualization.plot_connections(network=pn,
    ...                                       throats=Ts,
    ...                                       ax=ax,
    ...                                       color='r')  # plot boundary throats in red

    """
    import matplotlib.pyplot as plt
    from matplotlib import cm
    from matplotlib import colors as mcolors
    from matplotlib.collections import LineCollection
    from mpl_toolkits.mplot3d import Axes3D
    from mpl_toolkits.mplot3d.art3d import Line3DCollection

    from openpnm.topotools import dimensionality

    Ts = network.Ts if throats is None else network._parse_indices(throats)
    dim = dimensionality(network)
    ThreeD = True if dim.sum() == 3 else False
    # Add a dummy axis for 1D networks
    if dim.sum() == 1:
        dim[np.argwhere(~dim)[0]] = True

    if "fig" in kwargs.keys():
        raise Exception("'fig' argument is deprecated, use 'ax' instead.")
    if ax is None:
        fig, ax = plt.subplots()
    else:
        # The next line is necessary if ax was created using plt.subplots()
        fig, ax = ax.get_figure(), ax.get_figure().gca()
    if ThreeD and ax.name != '3d':
        fig.delaxes(ax)
        ax = fig.add_subplot(111, projection='3d')

    # Collect coordinates
    Ps = np.unique(network['throat.conns'][Ts])
    X, Y, Z = network['pore.coords'][Ps].T
    xyz = network["pore.coords"][:, dim]
    P1, P2 = network["throat.conns"][Ts].T
    throat_pos = np.column_stack((xyz[P1], xyz[P2])).reshape((Ts.size, 2, dim.sum()))

    # Deal with optional style related arguments
    if 'c' in kwargs.keys():
        color = kwargs.pop('c')
    color = mcolors.to_rgb(color) + tuple([alpha])
    if isinstance(cmap, str):
        try:
            cmap = plt.colormaps.get_cmap(cmap)
        except AttributeError:
            cmap = plt.cm.get_cmap(cmap)
    # Override colors with color_by if given
    if color_by is not None:
        color_by = np.array(color_by, dtype=np.float16)
        if len(color_by) != len(Ts):
            color_by = color_by[Ts]
        if not np.all(np.isfinite(color_by)):
            color_by[~np.isfinite(color_by)] = 0
            logger.warning('nans or infs found in color_by array, setting to 0')
        vmin = kwargs.pop('vmin', color_by.min())
        vmax = kwargs.pop('vmax', color_by.max())
        cscale = (color_by - vmin) / (vmax - vmin)
        color = cmap(cscale)
        color[:, 3] = alpha
    if size_by is not None:
        if len(size_by) != len(Ts):
            size_by = size_by[Ts]
        if not np.all(np.isfinite(size_by)):
            size_by[~np.isfinite(size_by)] = 0
            logger.warning('nans or infs found in size_by array, setting to 0')
        linewidth = size_by / size_by.max() * linewidth
    if label_by is not None:
        if len(label_by) != len(Ts):
            label_by = label_by[Ts]
    fontkws = kwargs.pop('font', {})

    if ThreeD:
        lc = Line3DCollection(throat_pos, colors=color, cmap=cmap,
                              linestyles=linestyle, linewidths=linewidth,
                              antialiaseds=np.ones_like(network.Ts), **kwargs)
    else:
        lc = LineCollection(throat_pos, colors=color, cmap=cmap,
                            linestyles=linestyle, linewidths=linewidth,
                            antialiaseds=np.ones_like(network.Ts), **kwargs)
        if label_by is not None:
            for count, (P1, P2) in enumerate(network.conns[Ts, :]):
                i, j, k = np.mean(network.coords[[P1, P2], :], axis=0)
                ax.text(i, j, label_by[count],
                        ha='center', va='center',
                        **fontkws)
    ax.add_collection(lc)

    if np.size(Ts) > 0:
        _scale_axes(ax=ax, X=X, Y=Y, Z=Z)
        _label_axes(ax=ax, X=X, Y=Y, Z=Z)
        fig.tight_layout()

    return lc


def plot_coordinates(network,
                     pores=None,
                     ax=None,
                     size_by=None,
                     color_by=None,
                     label_by=None,
                     cmap='jet',
                     color='r',
                     alpha=1.0,
                     marker='o',
                     markersize=10,
                     **kwargs):  # pragma: no cover
    r"""
    Produce a 3D plot showing specified pore coordinates as markers.

    Parameters
    ----------
    network : Network
        The network whose topological connections to plot.
    pores : array_like (optional)
        The list of pores to plot if only a sub-sample is desired. This is
        useful for inspecting a small region of the network. If no pores
        are specified then all are shown.
    ax : Matplotlib axis handle
        If ``ax`` is supplied, then the coordinates will be overlaid.
        This enables the plotting of multiple different sets of pores as
        well as throat connections from ``plot_connections``.
    size_by : str or array_like
        An ndarray of pore values (e.g. alg['pore.concentration']). These
        values are normalized by scaled by ``markersize``.  Note that this controls
        the marker *area*, so if you want the markers to be proportional to diameter
        you should do `size_by=net['pore.diameter']**2`.
    color_by : str or array_like
        An ndarray of pore values (e.g. alg['pore.concentration']).
    label_by : array_like (optional)
        An array or list of values to use as labels
    cmap : str or cmap object
        The matplotlib colormap to use if specfying a pore property
        for ``color_by``
    color : str
        A matplotlib named color (e.g. 'r' for red).
    alpha : float
        The transparency of the lines, with 1 being solid and 0 being invisible
    marker : 's'
        The marker to use.  The default is a circle.  Options are explained
        `here <https://matplotlib.org/3.2.1/api/markers_api.html>`_
    markersize : scalar
        Controls size of marker, default is 1.0.  This value is used to scale
        the ``size_by`` argument if given.
    font : dict
        A dictionary of key-value pairs that are used to control the font
        appearance if `label_by` is provided.
    **kwargs
        All other keyword arguments are passed on to the ``scatter``
        function of matplotlib, so check their documentation for additional
        formatting options.

    Returns
    -------
    pc : PathCollection
        Matplotlib object containing the markers representing the pores.

    Notes
    -----
    To create a single plot containing both pore coordinates and throats,
    consider creating an empty figure and then pass the ``ax`` object as
    an argument to ``plot_connections`` and ``plot_coordinates``.
    Otherwise, each call to either of these methods creates a new figure.

    See Also
    --------
    plot_connections

    Examples
    --------
    >>> import openpnm as op
    >>> import matplotlib as mpl
    >>> import matplotlib.pyplot as plt
    >>> mpl.use('Agg')
    >>> pn = op.network.Cubic(shape=[10, 10, 3])
    >>> pn['pore.internal'] = True
    >>> pn.add_boundary_pores()
    >>> Ps = pn.pores('internal')  # find internal pores
    >>> fig, ax = plt.subplots()  # create empty figure
    >>> _ = op.visualization.plot_coordinates(network=pn,
    ...                                       pores=Ps,
    ...                                       color='b',
    ...                                       ax=ax)  # plot internal pores
    >>> Ps = pn.pores('*boundary')  # find boundary pores
    >>> _ = op.visualization.plot_coordinates(network=pn,
    ...                                       pores=Ps,
    ...                                       color='r',
    ...                                       ax=ax)  # plot boundary pores in red

    """
    import matplotlib.pyplot as plt
    from matplotlib import cm
    from mpl_toolkits.mplot3d import Axes3D

    from openpnm.topotools import dimensionality

    Ps = network.Ps if pores is None else network._parse_indices(pores)

    dim = dimensionality(network)
    ThreeD = True if dim.sum() == 3 else False
    # Add a dummy axis for 1D networks
    if dim.sum() == 1:
        dim[np.argwhere(~dim)[0]] = True
    # Add 2 dummy axes for 0D networks (1 pore only)
    if dim.sum() == 0:
        dim[[0, 1]] = True

    if "fig" in kwargs.keys():
        raise Exception("'fig' argument is deprecated, use 'ax' instead.")
    if ax is None:
        fig, ax = plt.subplots()
    else:
        # The next line is necessary if ax was created using plt.subplots()
        fig, ax = ax.get_figure(), ax.get_figure().gca()
    if ThreeD and ax.name != '3d':
        fig.delaxes(ax)
        ax = fig.add_subplot(111, projection='3d')

    # Collect specified coordinates
    X, Y, Z = network['pore.coords'][Ps].T
    # The bounding box for fig is the entire ntwork (to fix the problem with
    # overwriting figures' axes lim)
    Xl, Yl, Zl = network['pore.coords'].T

    # Parse formatting kwargs
    if 'c' in kwargs.keys():
        color = kwargs.pop('c')
    if 's' in kwargs.keys():
        markersize = kwargs.pop('s')
    if isinstance(cmap, str):
        try:
            cmap = plt.colormaps.get_cmap(cmap)
        except AttributeError:
            cmap = plt.cm.get_cmap(cmap)
    if color_by is not None:
        color_by = np.array(color_by, dtype=np.float16)
        if len(color_by) != len(Ps):
            color_by = color_by[Ps]
        if not np.all(np.isfinite(color_by)):
            color_by[~np.isfinite(color_by)] = 0
            logger.warning('nans or infs found in color_by array, setting to 0')
        vmin = kwargs.pop('vmin', color_by.min())
        vmax = kwargs.pop('vmax', color_by.max())
        cscale = (color_by - vmin) / (vmax - vmin)
        color = cmap(cscale)
    if size_by is not None:
        if len(size_by) != len(Ps):
            size_by = size_by[Ps]
        if not np.all(np.isfinite(size_by)):
            size_by[~np.isfinite(size_by)] = 0
            logger.warning('nans or infs found in size_by array, setting to 0')
        markersize = size_by / size_by.max() * markersize
    if label_by is not None:
        if len(label_by) != len(Ps):
            label_by = label_by[Ps]
    fontkws = kwargs.pop('font', {})
    if ThreeD:
        sc = ax.scatter(X, Y, Z,
                        c=color,
                        s=markersize,
                        marker=marker,
                        alpha=alpha,
                        **kwargs)
        _scale_axes(ax=ax, X=Xl, Y=Yl, Z=Zl)
    else:
        _X, _Y = np.column_stack((X, Y, Z))[:, dim].T
        sc = ax.scatter(_X, _Y,
                        c=color,
                        s=markersize,
                        marker=marker,
                        alpha=alpha,
                        **kwargs)
        if label_by is not None:
            for count, (i, j, k) in enumerate(network.coords[Ps, :]):
                ax.text(i, j, label_by[count],
                        ha='center', va='center',
                        **fontkws)
        _scale_axes(ax=ax, X=Xl, Y=Yl, Z=np.zeros_like(Yl))

    _label_axes(ax=ax, X=Xl, Y=Yl, Z=Zl)
    fig.tight_layout()

    return sc


def _label_axes(ax, X, Y, Z):
    labels = ["X", "Y", "Z"]
    dim = np.zeros(3, dtype=bool)
    for i, arr in enumerate([X, Y, Z]):
        if np.unique(arr).size > 1:
            dim[i] = True
    # Add a dummy axis for 1D networks
    if dim.sum() == 1:
        dim[np.argwhere(~dim)[0]] = True
    # Add 2 dummy axes for 0D networks (1 pore only)
    if dim.sum() == 0:
        dim[[0, 1]] = True
    dim_idx = np.argwhere(dim).squeeze()
    ax.set_xlabel(labels[dim_idx[0]])
    ax.set_ylabel(labels[dim_idx[1]])
    if hasattr(ax, "set_zlim"):
        ax.set_zlabel("Z")


def _scale_axes(ax, X, Y, Z):
    max_range = np.ptp([X, Y, Z]).max() / 2
    mid_x = (X.max() + X.min()) * 0.5
    mid_y = (Y.max() + Y.min()) * 0.5
    mid_z = (Z.max() + Z.min()) * 0.5
    ax.set_xlim(mid_x - max_range, mid_x + max_range)
    ax.set_ylim(mid_y - max_range, mid_y + max_range)
    if hasattr(ax, "set_zlim"):
        ax.set_zlim(mid_z - max_range, mid_z + max_range)
    else:
        ax.axis("equal")


def plot_networkx(network,
                  plot_throats=True,
                  labels=None,
                  colors=None,
                  scale=1,
                  ax=None,
                  alpha=1.0):  # pragma: no cover
    r"""
    Creates a pretty 2D plot for 2D OpenPNM networks.

    Parameters
    ----------
    network : Network
    plot_throats : bool, optional
        Plots throats as well as pores, if True.
    labels : list, optional
        List of OpenPNM labels
    colors : list, optional
        List of corresponding colors to the given `labels`.
    scale : float, optional
        Scale factor for size of pores.
    ax : matplotlib.Axes, optional
        Matplotlib axes object
    alpha: float, optional
        Transparency value, 1 is opaque and 0 is transparent

    """
    import matplotlib.pyplot as plt
    from matplotlib.collections import PathCollection
    from networkx import Graph, draw_networkx_edges, draw_networkx_nodes

    from openpnm.topotools import dimensionality

    dims = dimensionality(network)
    if dims.sum() > 2:
        raise Exception("NetworkX plotting only works for 2D networks.")
    temp = network['pore.coords'].T[dims].squeeze()
    if dims.sum() == 1:
        x = temp
        y = np.zeros_like(x)
    if dims.sum() == 2:
        x, y = temp
    try:
        node_size = scale * network['pore.diameter']
    except KeyError:
        node_size = np.ones_like(x) * scale * 0.5
    G = Graph()
    pos = {network.Ps[i]: [x[i], y[i]] for i in range(network.Np)}
    if not np.isfinite(node_size).all():
        node_size[~np.isfinite(node_size)] = np.nanmin(node_size)
    node_color = np.array(['k'] * len(network.Ps))

    if labels:
        if not isinstance(labels, list):
            labels = [labels]
        if not isinstance(colors, list):
            colors = [colors]
        if len(labels) != len(colors):
            raise Exception('len(colors) must be equal to len(labels)!')
        for label, color in zip(labels, colors):
            node_color[network.pores(label)] = color

    if ax is None:
        fig, ax = plt.subplots()
    ax.set_aspect('equal', adjustable='datalim')
    offset = node_size.max() * 0.5
    ax.set_xlim((x.min() - offset, x.max() + offset))
    ax.set_ylim((y.min() - offset, y.max() + offset))
    ax.axis("off")

    # Keep track of already plotted nodes
    temp = [id(item) for item in ax.collections if isinstance(item, PathCollection)]

    # Plot pores
    gplot = draw_networkx_nodes(G, ax=ax, pos=pos, nodelist=network.Ps.tolist(),
                                alpha=alpha, node_color=node_color, edgecolors=node_color,
                                node_size=node_size)
    # (Optionally) Plot throats
    if plot_throats:
        draw_networkx_edges(G, pos=pos, edge_color='k', alpha=alpha,
                            edgelist=network['throat.conns'].tolist(), ax=ax)

    spi = 2700  # 1250 was obtained by trial and error
    figwidth, figheight = ax.get_figure().get_size_inches()
    figsize_ratio = figheight / figwidth
    data_ratio = ax.get_data_ratio()
    corr = min(figsize_ratio / data_ratio, 1)
    xrange = np.ptp(ax.get_xlim())
    markersize = np.atleast_1d((corr*figwidth)**2 / xrange**2 * node_size**2 * spi)
    for item in ax.collections:
        if isinstance(item, PathCollection) and id(item) not in temp:
            item.set_sizes(markersize)

    return gplot


def plot_tutorial(network,
                  pore_labels=None,
                  throat_labels=None,
                  font_size=12,
                  line_width=2,
                  node_color='b',
                  edge_color='r',
                  node_size=500):  # pragma: no cover
    r"""
    Generate a network plot suitable for tutorials and explanations.

    Parameters
    ----------
    network : Network
        The network to plot, should be 2D, since the z-coordinate will be
        ignored.
    pore_labels : array_like
        A list of values to use for labeling the pores. If not provided then pore
        index is used.
    throat_labels : array_like
        A list of values to use for labeling the throat. If not provided then throat
        index is used.
    font_size : int
        Size of font to use for labels.
    line_width : int
        Thickness of edge lines and node borders.
    node_color : str
        Color of node border.
    edge_color : str
        Color of edge lines.
    node_size : int
        Size of node circle.

    Returns
    -------
    g : NetworkX plot object

    """
    import matplotlib.pyplot as plt
    import networkx as nx

    from openpnm.io import network_to_networkx

    G = network_to_networkx(network=network)
    pos = {i: network['pore.coords'][i, 0:2] for i in network.Ps}
    if pore_labels is None:
        labels = {i: i for i in network.Ps}
    else:
        labels = {i: pore_labels[i] for i in network.Ps}
    if throat_labels is None:
        edge_labels = {tuple(network['throat.conns'][i, :]): i for i in network.Ts}
    else:
        edge_labels = {tuple(network['throat.conns'][i, :]): throat_labels[i]
                       for i in network.Ts}

    gplot = nx.draw_networkx_nodes(G, pos,
                                   node_size=node_size,
                                   node_color='w',
                                   edgecolors=node_color,
                                   linewidths=line_width)
    nx.draw_networkx_edges(
        G, pos, width=line_width, edge_color=edge_color)
    nx.draw_networkx_labels(
        G, pos, labels=labels, font_size=font_size, font_color='k')
    nx.draw_networkx_edge_labels(
        G, pos, edge_labels=edge_labels, font_size=font_size, font_color='k')

    # Prettify the figure (margins, etc.)
    plt.axis('off')
    ax = plt.gca()
    ax.margins(0.1, 0.1)
    ax.set_aspect("equal")
    fig = plt.gcf()
    fig.tight_layout()
    dims = op.topotools.dimensionality(network)
    xy_range = np.ptp(network.coords, axis=0)[dims]
    aspect_ratio = xy_range[0] / xy_range[1]
    fig.set_size_inches(5, 5 / aspect_ratio)

    return gplot


def plot_notebook(network,
                  node_color=0,
                  edge_color=0,
                  node_size=1,
                  node_scale=20,
                  edge_scale=5,
                  colormap='viridis'):
    r"""
    Visualize a network in 3D using Plotly.

    The pores and throats are scaled and colored by their properties.
    The final figure can be rotated and zoomed.

    Parameters
    ----------
    network : Network
        The network to visualize
    node_color : ndarray
        An array of values used for coloring the pores. If not given, the
        lowest value of the employed colormap is assigned to all markers.
    edge_color : ndarray
        An array of values used for coloring the throats. If not given, the
        lowest value of the employed colormap is assigned to all lines.
    node_size : ndarray
        An array of values controlling the size of the markers.  If not given
        all markers will be the same size
    node_scale : scalar
        A scaler to resize the markers
    edge_scale : scalar
        A scaler to the line thickness
    colormap : str
        The colormap to use

    Returns
    -------
    fig : Plotly graph object
        The graph object containing the generated plots. The object has
        several useful methods.

    Notes
    -----
    **Important**

    a) This does not work in Spyder. It should only be called from a
    Jupyter Notebook.

    b) This is only meant for relatively small networks. For proper
    visualization use Paraview.

    """
    try:
        import plotly.graph_objects as go
    except ImportError:
        raise Exception('Plotly is not installed.'
                        'Please install Plotly using "pip install plotly"')

    # Get xyz coords for points
    x_nodes, y_nodes, z_nodes = network.coords.T

    node_size = np.ones(network.Np)*node_size
    node_color = np.ones(network.Np)*node_color
    edge_color = np.ones(network.Nt)*edge_color

    node_labels = [str(i) + ': ' + str(x) for i, x in
                   enumerate(zip(node_size, node_color))]
    edge_labels = [str(i) + ': ' + str(x) for i, x in enumerate(edge_color)]

    # Create edges and nodes coordinates
    N = network.Nt*3

    x_edges = np.zeros(N)
    x_edges[np.arange(0, N, 3)] = network.coords[network.conns[:, 0]][:, 0]
    x_edges[np.arange(1, N, 3)] = network.coords[network.conns[:, 1]][:, 0]
    x_edges[np.arange(2, N, 3)] = np.nan

    y_edges = np.zeros(network.Nt*3)
    y_edges[np.arange(0, N, 3)] = network.coords[network.conns[:, 0]][:, 1]
    y_edges[np.arange(1, N, 3)] = network.coords[network.conns[:, 1]][:, 1]
    y_edges[np.arange(2, N, 3)] = np.nan

    z_edges = np.zeros(network.Nt*3)
    z_edges[np.arange(0, N, 3)] = network.coords[network.conns[:, 0]][:, 2]
    z_edges[np.arange(1, N, 3)] = network.coords[network.conns[:, 1]][:, 2]
    z_edges[np.arange(2, N, 3)] = np.nan

    # Create plotly's Scatter3d object for pores and throats
    trace_edges = go.Scatter3d(x=x_edges,
                               y=y_edges,
                               z=z_edges,
                               mode='lines',
                               line=dict(color=edge_color,
                                         width=edge_scale,
                                         colorscale=colormap),
                               text=edge_labels, hoverinfo='text')

    trace_nodes = go.Scatter3d(x=x_nodes,
                               y=y_nodes,
                               z=z_nodes,
                               mode='markers',
                               marker=dict(symbol='circle',
                                           size=node_size*node_scale,
                                           color=node_color,
                                           colorscale=colormap,
                                           line=dict(color='black', width=0.5)),
                               text=node_labels, hoverinfo='text')

    axis = dict(showbackground=False,
                showline=False,
                zeroline=False,
                showgrid=False,
                showticklabels=False,
                title='')

    layout = go.Layout(width=650,
                       height=625,
                       showlegend=False,
                       scene=dict(xaxis=dict(axis),
                                  yaxis=dict(axis),
                                  zaxis=dict(axis),),
                       margin=dict(t=100),
                       hovermode='closest')

    data = [trace_edges, trace_nodes]
    fig = go.Figure(data=data, layout=layout)
    return fig


def create_pore_colors_from_array(a, cmap='viridis'):
    colormap = cm.get_cmap(cmap)
    return colormap(a/a.max())


def create_throat_colors_from_array(a, cmap='viridis'):
    colormap = cm.get_cmap(cmap)
    return np.repeat(colormap(a/a.max()), 2, axis=0)


def plot_vispy(
    network,
    pore_color=None,
    pore_size=None,
    throat_color=None,
    throat_size=None,
    bgcolor='grey',
):
    r"""
    Creates a pretty network plot using VisPy.

    Parameters
    ----------
    network
    """
    try:
        from vispy import scene
    except ModuleNotFoundError:
        raise Exception("vispy must be installed to use this function")
    canvas = scene.SceneCanvas(keys='interactive', show=True, bgcolor=bgcolor)
    view = canvas.central_widget.add_view()
    view.camera = 'turntable'
    view.camera.fov = 30
    view.camera.distance = 3*np.max(network['pore.coords'])

    if pore_color is None:
        pore_color = create_pore_colors_from_array(network['pore.diameter'],
                                                   cmap='viridis')
    else:
        pore_color = create_pore_colors_from_array(pore_color,
                                                   cmap='viridis')
    if throat_color is None:
        throat_color = create_throat_colors_from_array(network['throat.diameter'],
                                                       cmap='viridis')
    else:
        throat_color = create_throat_colors_from_array(throat_color,
                                                       cmap='viridis')
    if pore_size is None:
        pore_size = network['pore.diameter']
    if throat_size is None:
        throat_size = 2
    else:
        throat_size = np.max(throat_size)  # Arrays not supported here
    # plot spheres
    vis = scene.visuals.Markers(
        pos=network['pore.coords'],
        size=pore_size,
        antialias=0,
        face_color=pore_color,
        edge_width=0,
        scaling=True,
        spherical=True,
    )
    vis.parent = view.scene
    # plot axis
    # vispy.scene.visuals.XYZAxis(parent=view.scene)
    # set camera center
    view.camera.center = np.array((network['pore.coords'][:, 0].max()/2,
                                   network['pore.coords'][:, 1].max()/2,
                                   network['pore.coords'][:, 2].max()/2))
    # data preparation
    lines = np.zeros((len(network['throat.conns']), 2, 3))
    for i in range(len(network['throat.conns'])):
        pair = network['throat.conns'][i]
        line = np.array([[network['pore.coords'][pair[0]],
                          network['pore.coords'][pair[1]]]])
        lines[i, :, :] = line
    # plot throats
    vis2 = scene.visuals.Line(lines,
                              width=throat_size,
                              color=throat_color,
                              connect='segments',
                              antialias=True,)
    vis2.parent = view.scene


def set_mpl_style():  # pragma: no cover
    r"""
    Prettifies matplotlib's output by adjusting fonts, markersize etc.
    """
    sfont = 12
    mfont = 12
    lfont = 12

    image_props = {'interpolation': 'none',
                   'cmap': 'viridis'}
    line_props = {'linewidth': 2,
                  'markersize': 7,
                  'markerfacecolor': 'w'}
    font_props = {'size': sfont}
    axes_props = {'titlesize': lfont,
                  'labelsize': mfont,
                  'linewidth': 2,
                  'labelpad': 8}
    xtick_props = {'labelsize': sfont,
                   'top': True,
                   'direction': 'in',
                   'major.size': 6,
                   'major.width': 2}
    ytick_props = {'labelsize': sfont,
                   'right': True,
                   'direction': 'in',
                   'major.size': 6,
                   'major.width': 2}
    legend_props = {'fontsize': mfont,
                    'frameon': False}
    figure_props = {'titlesize': sfont,
                    'autolayout': True}

    plt.rc('font', **font_props)
    plt.rc('lines', **line_props)
    plt.rc('axes', **axes_props)
    plt.rc('xtick', **xtick_props)
    plt.rc('ytick', **ytick_props)
    plt.rc('legend', **legend_props)
    plt.rc('figure', **figure_props)
    plt.rc('image', **image_props)

    import matplotlib_inline
    matplotlib_inline.backend_inline.set_matplotlib_formats('retina')



import logging
import pickle
import re
import uuid
from copy import deepcopy
from datetime import datetime

import numpy as np

from openpnm.utils import SettingsAttr, Workspace

ws = Workspace()
logger = logging.getLogger(__name__)


__all__ = [
    'Project',
]


class ProjectSettings(SettingsAttr):
    r"""
    uuid : str
        A universally unique identifier for the object to keep things straight
    """
    uuid = ''
    original_uuid = ''

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        name = ws._validate_name(name)
        for v in list(ws.values()):
            if v.settings is self:
                ws[name] = ws.pop(v.settings.name)
        self._name = name


class Project(list):
    r"""
    This class provides a container for all OpenPNM objects in a given
    simulation.

    A simulation is defined as a Network and all of it's associated objects.
    When instantiating a Network, a Project can be passed as an argument, but
    if not given one is created. When instantiating any other object either
    a Network or a Project can be supplied. In the former case, the
    Network's Project is retrieved and used. The end result is that all
    objects are stored in a specific Project.

    The Project to which any object belongs can be retrieved with
    ``obj.project``. Conversely, printing a Project displays a list of all
    objects it contains.

    Moreover, all Projects are registered with the Workspace. Since there can
    be only instance of the Workspace it is possible to view all open Projects
    by printing the Workspace.

    See Also
    --------
    Workspace

    """

    def __init__(self, *args, **kwargs):
        self.settings = ProjectSettings()
        self.settings['name'] = kwargs.pop('name', None)
        self.settings['uuid'] = str(uuid.uuid4())
        self.settings['original_uuid'] = self.settings['uuid']
        super().__init__(*args, **kwargs)
        ws[self.settings['name']] = self

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except TypeError:
            if isinstance(key, str):  # Enable dict-style lookup if key is a string
                for item in self:
                    if item.name == key:
                        return item
            raise KeyError(key)

    def copy(self, name=None):
        r"""
        Creates a deep copy of the current project

        A deep copy means that new, unique versions of all the objects are
        created but with identical data and properties.

        Parameters
        ----------
        name : str
            The name to give to the new project. If not supplied, a name
            is automatically generated.

        Returns
        -------
        proj : list
            A new Project object containing copies of all objects

        Notes
        -----
        Because they are new objects, they are given a new uuid
        (``obj.settings['uuid']``), but the uuid of the original object
        is also stored (``obj.settings['original_uuid']``) for reference.

        """
        name = ws._validate_name(name)
        proj = deepcopy(self)
        for item in proj:
            item.settings['uuid'] = str(uuid.uuid4())
        proj.settings['uuid'] = str(uuid.uuid4())
        proj.settings['name'] = name
        ws[name] = proj
        return proj

    def _set_name(self, name):
        self.settings['name'] = name

    def _get_name(self):
        return self.settings['name']

    name = property(fget=_get_name, fset=_set_name)

    def _validate_name(self, name):
        if name in self.names:
            raise Exception('Another object is already named '+name)
        for item in self:
            for key in item.keys():
                if key.split('.')[1] == name:
                    raise Exception('A property/label is already named '+name)

    def _generate_name(self, name=''):
        if name in [None, '']:
            name = 'obj_01'  # Give basic name, then let rest of func fix it
        warn = True
        if name.endswith('_?'):
            name = name.replace('_?', '_01')
            warn = False
        if name in self.names:  # If proposed name is taken, increment it
            proposed_name = name
            if not re.search(r'_\d+$', name):  # If name does not end with _##
                name = name + '_01'
            prefix, count = name.rsplit('_', 1)
            n = [0]
            for item in self:
                if item.name.startswith(prefix+'_'):
                    n.append(int(item.name.split(prefix+'_')[1]))
            name = prefix+'_'+str(max(n)+1).zfill(2)
            if warn:
                logger.warn(f'{proposed_name} is already taken, using {name} instead')
        self._validate_name(name)
        return name

    @property
    def names(self):
        names = [i.name for i in self]
        return names

    @property
    def network(self):
        for item in self:
            if ('throat.conns' in item.keys()) or ('pore.coords' in item.keys()):
                return item

    @property
    def phases(self):
        from openpnm.phase import Phase
        phases = []
        for item in self:
            if isinstance(item, Phase):
                phases.append(item)
        return phases

    @property
    def algorithms(self):
        from openpnm.algorithms import Algorithm
        algs = []
        for item in self:
            if isinstance(item, Algorithm):
                algs.append(item)
        return algs

    @property
    def workspace(self):
        return ws

    def _get_locations(self, label):
        r"""
        Find locations indicated by the given label regardless of which object
        it is defined on

        Parameters
        ----------
        label : str
            The label whose locations are sought, such as 'pore.left'

        Returns
        -------
        locations : ndarray
            A boolean array with ``True`` values indicating which locations
            have the given label

        Notes
        -----
        The returns the first instance of ``label`` that it finds
        """
        for item in self:
            try:
                return item[label]
            except KeyError:
                pass
        raise KeyError(label)

    def __str__(self):  # pragma: no cover
        hr = '―'*78
        s = '═'*78 + '\n'
        s += 'Object Name : Object Class and ID' + '\n'
        s += hr + '\n'
        for item in self:
            s += item.__repr__() + '\n'
        s += hr
        return s



from openpnm.models.collections.phase import water
from openpnm.phase import Phase, _fetch_chemical_props
from openpnm.utils import Docorator


docstr = Docorator()


__all__ = [
    'Water',
]


@docstr.dedent
class Water(Phase):
    r"""
    Creates Phase object with preset values for Water

    Parameters
    ----------
    %(Phase.parameters)s

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        from thermo import Chemical
        a = Chemical('h2o')
        temp = _fetch_chemical_props(a)
        self.params.update(temp)
        self.models.update(water)
        self.regenerate_models()



import openpnm.models as mods


standard_liquid = {
    'pore.density': {
        'model': mods.phase.density.liquid_pure_COSTALD,
    },
    'pore.heat_capacity_gas': {
        'model': mods.phase.heat_capacity.gas_pure_TRC,
    },
    'pore.heat_capacity': {
        'model': mods.phase.heat_capacity.liquid_pure_rp,
    },
    'pore.thermal_conductivity': {
        'model': mods.phase.thermal_conductivity.liquid_pure_gismr,
    },
    'pore.viscosity': {
        'model': mods.phase.viscosity.liquid_pure_ls,
    },
    'pore.vapor_pressure': {
        'model': mods.phase.vapor_pressure.liquid_pure_antoine,
    },
}


standard_liquid_mixture = {
    'pore.density': {
        'model': mods.phase.density.liquid_mixture_COSTALD,
    },
    'pore.heat_capacity': {
        'model': mods.phase.heat_capacity.liquid_mixture_xweighted,
    },
    'pore.thermal_conductivity': {
        'model': mods.phase.thermal_conductivity.liquid_mixture_DIPPR9H,
    },
    'pore.viscosity': {
        'model': mods.phase.viscosity.liquid_mixture_xweighted,
    },
}



from openpnm.models.physics._utils import _poisson_conductance
from openpnm.models import _doctxt


__all__ = ["generic_electrical", "series_resistors"]


@_doctxt
def generic_electrical(
    phase,
    pore_conductivity='pore.electrical_conductivity',
    throat_conductivity='throat.electrical_conductivity',
    size_factors='throat.diffusive_size_factors'
):
    r"""
    Calculate the electrical conductance of conduits in network, where a
    conduit is ( 1/2 pore - full throat - 1/2 pore ). See the notes section.

    Parameters
    ----------
    %(phase)s
    pore_conductivity : str
        %(dict_blurb)s pore electrical conductivity
    throat_conductivity : str
        %(dict_blurb)s throat electrical conductivity
    size_factors: str
        %(dict_blurb)s conduit diffusive size factors

    Returns
    -------
    %(return_arr)s electrical conductance

    """
    return _poisson_conductance(phase=phase,
                                pore_conductivity=pore_conductivity,
                                throat_conductivity=throat_conductivity,
                                size_factors=size_factors)


def series_resistors(
    phase,
    pore_conductivity='pore.electrical_conductivity',
    throat_conductivity='throat.electrical_conductivity',
    size_factors='throat.diffusive_size_factors'
):
    r"""
    Calculate the electrical conductance of conduits in network, where a
    conduit is ( 1/2 pore - full throat - 1/2 pore ). See the notes section.

    Parameters
    ----------
    %(phase)s
    pore_conductivity : str
        %(dict_blurb)s pore electrical conductivity
    throat_conductivity : str
        %(dict_blurb)s throat electrical conductivity
    size_factors: str
        %(dict_blurb)s conduit diffusive size factors

    Returns
    -------
    %(return_arr)s electrical conductance

    """
    return _poisson_conductance(phase=phase,
                                pore_conductivity=pore_conductivity,
                                throat_conductivity=throat_conductivity,
                                size_factors=size_factors)



import logging
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import Docorator


logger = logging.getLogger(__name__)
docstr = Docorator()


__all__ = ['FourierConduction']


@docstr.dedent
class FourierConductionSettings:
    r'''

    Parameters
    ----------
    %(ReactiveTransportSettings.parameters)s

    '''
    quantity = 'pore.temperature'
    conductance = 'throat.thermal_conductance'


class FourierConduction(ReactiveTransport):
    r"""
    A subclass of LinearTransport to simulate heat conduction

    """

    def __init__(self, name='fourier_?', **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(FourierConductionSettings())



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PatchCollection
from matplotlib.patches import Circle, Polygon, Rectangle

from openpnm._skgraph.tools import rotate_coords
from openpnm.models.geometry import throat_endpoints, throat_length
from openpnm.network import Network
from openpnm.topotools import find_connected_sites

__all__ = [
    'draw_conduit',
]


def draw_conduit(network, throat):
    """Draws a subset of a network given throat numbers."""
    pn = network
    P1, P2 = find_connected_sites(g=pn, bonds=throat)
    new_net = Network(coords=pn.coords[[P1, P2], :], conns=np.atleast_2d([0, 1]))
    new_net.regenerate_models()
    new_net['pore.diameter'] = pn['pore.diameter'][[P1, P2]]
    new_net['throat.diameter'] = [pn['throat.diameter'][throat]]
    new_net['throat.length'] = throat_length.circles_and_rectangles(new_net)
    new_net['throat.endpoints'] = throat_endpoints.spheres_and_cylinders(new_net)

    Pcrds = np.vstack((pn.coords[0, :], pn.coords[0, :]))
    Pcrds[1, :] = Pcrds[1, :] + np.array([new_net['throat.spacing'], 0, 0])
    fig, ax = plt.subplots(figsize=[10, 5])
    patches = []
    for i, P in enumerate([0, 1]):
        x, y, r = Pcrds[P, 0], Pcrds[P, 1], new_net['pore.diameter'][P]/2
        circle = Circle((x, y), r)
        patches.append(circle)
    Tcrds = np.vstack(list(new_net['throat.endpoints'].values()))
    Tcrds = rotate_coords(Tcrds, b=270)
    H = new_net['throat.diameter'][0]
    W = new_net['throat.length'][0]
    R1 = new_net['pore.diameter'][0]/2
    R2 = new_net['pore.diameter'][1]/2
    rect = Rectangle(xy=(Tcrds[0, 0], Tcrds[0, 1]-H/2), height=H, width=W)
    patches.append(rect)
    p = PatchCollection(patches, alpha=0.5, edgecolor='k', linewidth=3)
    p.set_array([1, 1, 2])
    p.cmap = plt.cm.bwr
    ax.add_collection(p)
    ax.scatter(*Pcrds[0, :2], marker='+', c='k', s=100)
    ax.scatter(*Pcrds[1, :2], marker='+', c='k', s=100)
    left = min(Pcrds[:, 0])-R1
    right = max(Pcrds[:, 0])+R2
    ax.set_xlim([left - abs(left-right)/32, right + abs(left-right)/32])
    temp = (Pcrds[:, 1]).mean()
    ax.set_ylim([temp - abs(left-right)/4, temp + abs(left-right)/4])
    ax.annotate(text='',
                xy=Pcrds[0, :2],
                xytext=(Pcrds[0, 0], Pcrds[0, 1]-R1),
                arrowprops=dict(arrowstyle='<-', lw=2))
    ax.annotate(text=f"{R1}",
                xy=Pcrds[0, :2],
                xytext=(Pcrds[0, 0], Pcrds[0, 1]-R1/2),
                textcoords='data')
    return ax



try:
    from matplotlib._docstring import Substitution
except ModuleNotFoundError:
    from matplotlib.docstring import Substitution


__all__ = [
    '_geodocs',
]


_geodocs = Substitution(
    network=
    r"""network : OpenPNM Network object

    """,
    Dp=
    r"""pore_diameter : str
            Name of the dictionary key on ``network`` containing the ndarray of
            pore diameter values.""",
    Dt=
    r"""throat_diameter : str
            Name of the dictionary key on ``network`` containing the ndarray of
            throat diameter values.""",
    Vp=
    r"""pore_volume : str
            Name of the dictionary key on ``network`` containing the ndarray of
            pore volume values.""",
    Vt=
    r"""throat_volume : str
            Name of the dictionary key on ``network`` containing the ndarray of
            throat volume values.""",
    Lt=
    r"""throat_length : str
            Name of the dictionary key on ``network`` containing the ndarray of
            throat length values.""",
    Pcoords=
    r"""pore_coords : str
            Name of the dictionary key on ``network`` containing the ndarray of
            pore coordinate values.""",
    Tcoords=
    r"""throat_coords : str
            Name of the dictionary key on ``network`` containing the ndarray of
            throat centroid coordinate values.""",
    At=
    r"""throat_area : str
            Name of the dictionary key on ``network`` containing the ndarray of
            throat surface values.""",
    Pt=
    r"""throat_perimeter : str
            Name of the dictionary key on ``network`` containing the ndarray of
            throat perimeter values.""",
    Tcen=
    r"""throat_centroid : str
            Name of the dictionary key on ``network`` containing the ndarray of
            throat centroid coordinate values.""",
    Act=
    r"""throat_cross_sectional_area : str
            Name of the dictionary key on ``network`` containing the ndarray of
            throat cross-sectional area values.""",
)



from openpnm.models.collections.phase import air
from openpnm.phase import Phase, _fetch_chemical_props
from openpnm.phase import StandardGas, StandardGasMixture
from openpnm.utils import Docorator


docstr = Docorator()


__all__ = [
    'Air',
    '_AirMixture',
]


@docstr.dedent
class Air(Phase):
    r"""
    Creates a Phase object with preset models and values for air

    Parameters
    ----------
    %(Phase.parameters)s

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        from thermo import Mixture
        a = Mixture(IDs=['O2', 'N2'], zs=[0.21, 0.79])
        temp = _fetch_chemical_props(a)
        self.params.update(temp)
        self.models.update(air)
        self.regenerate_models()


class _AirMixture(StandardGasMixture):  # pragma: no cover

    def __init__(self, network, **kwargs):
        o2 = StandardGas(network=network, species='O2')
        n2 = StandardGas(network=network, species='N2')
        super().__init__(network=network, components=[o2, n2], **kwargs)
        self.y(o2, 0.21)
        self.y(n2, 0.79)
        self.regenerate_models()



import numpy as np
import scipy.spatial as sptl
from openpnm._skgraph.generators import delaunay as _delaunay


__all__ = [
    'gabriel',
]


def gabriel(points=None, delaunay=None, shape=None, reflect=False,
            node_prefix='node', edge_prefix='edge'):
    r"""
    Generate a network based on a Gabriel tessellation, which is a subset of
    the Delaunay triangulation

    Parameters
    ----------
    points : array_like or int, optional
        Can either be an N-by-3 array of point coordinates which will be used,
        or a scalar value indicating the number of points to generate.
        This can be omitted if ``delaunay`` is provided.
    delaunay : network dictionary, optional
        A dictionary containing coords and conns as produced by the
        ``delaunay`` function.  If ``points`` are provided this is
        ignored.
    shape : array_like
        Indicates the size and shape of the domain
    reflect : boolean, optional (default = ``False``)
        If ``True`` then points are reflected across each face of the domain
        prior to performing the tessellation. These reflected points are
        automatically trimmed.  Enabling this behavior prevents long-range
        connections between surface pores.

    Returns
    -------
    network : dict
        A dictionary containing 'node.coords' and 'edge.conns'

    """
    if points is not None:
        dn, tri = _delaunay(points=points, shape=shape, reflect=reflect,
                            node_prefix=node_prefix, edge_prefix=edge_prefix)
    else:
        dn = delaunay
    # Find centroid or midpoint of each edge in conns
    c = dn[node_prefix+'.coords'][dn[edge_prefix+'.conns']]
    m = (c[:, 0, :] + c[:, 1, :])/2
    # Find the radius the sphere between each pair of nodes
    r = np.sqrt(np.sum((c[:, 0, :] - c[:, 1, :])**2, axis=1))/2
    # Use the kd-tree function in Scipy's spatial module
    tree = sptl.cKDTree(dn[node_prefix+'.coords'])
    # Find the nearest point for each midpoint
    n = tree.query(x=m, k=1)[0]
    # If nearest point to m is at distance r, then the edge is a Gabriel edge
    g = n >= r*(0.999)  # This factor avoids precision errors in the distances
    d = {}
    d.update(dn)
    # Reduce the connectivity to all True values found in g
    d[edge_prefix+'.conns'] = dn[edge_prefix+'.conns'][g]
    d[node_prefix+'.coords'] = dn[node_prefix+'.coords']
    return d



import numpy as np
import scipy.spatial as sptl
import scipy.sparse as sprs
from numba import njit
from openpnm._skgraph.generators import cubic
from openpnm._skgraph.tools import tri_to_am


@njit
def len_lil(lil):
    indptr = [len(i) for i in lil]
    return indptr


def fcc(shape, spacing=1, mode='kdtree', node_prefix='node', edge_prefix='edge'):
    r"""
    Generate a face-centered cubic lattice

    Parameters
    ----------
    shape : array_like
        The number of corner sites in each direction. A sipmle cubic lattice
        is created then the 'face-sites' are added afterwards.
    spacing : array_like or float
        The size of a unit cell in each direction. If an scalar is given it is
        applied in all 3 directions.
    mode : str
        Dictate how neighbors are found.  Options are:

        ===============  =====================================================
        mode             meaning
        ===============  =====================================================
        'kdtree'         Uses ``scipy.spatial.KDTree`` to find all neighbors
                         within the unit cell.
        'triangulation'  Uses ``scipy.spatial.Delaunay`` to find all neighbors.
        ===============  =====================================================

    Returns
    -------
    network : dict
        A dictionary containing 'coords', 'conns' and various boolean labels
        (i.e. 'node.center')

    Notes
    -----
    It is not clear whether KDTree of Delaunay are faster. In fact it is
    surely possible to find the neighbors formulaically but this is not
    implemented yet.

    """
    shape = np.array(shape)
    # Create base cubic network of corner sites
    net1 = cubic(shape=shape,
                 node_prefix=node_prefix, edge_prefix=edge_prefix)
    # Create 3 networks to become face sites
    net2 = cubic(shape=shape - [1, 1, 0],
                 node_prefix=node_prefix, edge_prefix=edge_prefix)
    net3 = cubic(shape=shape - [1, 0, 1],
                 node_prefix=node_prefix, edge_prefix=edge_prefix)
    net4 = cubic(shape=shape - [0, 1, 1],
                 node_prefix=node_prefix, edge_prefix=edge_prefix)
    # Offset pore coords by 1/2 a unit cell
    net2[node_prefix+'.coords'] += np.array([0.5, 0.5, 0])
    net3[node_prefix+'.coords'] += np.array([0.5, 0, 0.5])
    net4[node_prefix+'.coords'] += np.array([0, 0.5, 0.5])
    crds = np.concatenate((net1[node_prefix+'.coords'],
                           net2[node_prefix+'.coords'],
                           net3[node_prefix+'.coords'],
                           net4[node_prefix+'.coords']))
    corner_labels = np.concatenate(
        (np.ones(net1[node_prefix+'.coords'].shape[0], dtype=bool),
         np.zeros(net2[node_prefix+'.coords'].shape[0], dtype=bool),
         np.zeros(net3[node_prefix+'.coords'].shape[0], dtype=bool),
         np.zeros(net4[node_prefix+'.coords'].shape[0], dtype=bool)))
    if mode.startswith('tri'):
        tri = sptl.Delaunay(points=crds)
        am = tri_to_am(tri)
        conns = np.vstack((am.row, am.col)).T
        # Trim diagonal connections between cubic pores
        L = np.sqrt(np.sum(np.diff(crds[conns], axis=1)**2, axis=2)).flatten()
        conns = conns[L <= 0.75]
    elif mode.startswith('kd'):
        tree1 = sptl.KDTree(crds)
        # Method 1
        hits = tree1.query_ball_point(crds, r=.75)
        # Method 2: Not sure which is faster
        # tree2 = sptl.KDTree(crds)
        # hits = tree1.query_ball_tree(tree1, r=1)
        indices = np.hstack(hits)
        # Convert to CSR matrix
        indptr = [len(i) for i in hits]
        indptr.insert(0, 0)
        indptr = np.cumsum(indptr)
        am = sprs.csr_matrix((np.ones_like(indices), indices, indptr))
        am = sprs.triu(am, k=1)
        am = am.tocoo()
        conns = np.vstack((am.row, am.col)).T
    conns = np.vstack((net1[edge_prefix+'.conns'], conns))

    d = {}
    d[node_prefix + '.coords'] = crds*spacing
    d[edge_prefix + '.conns'] = conns
    d[node_prefix + '.corner'] = corner_labels
    d[node_prefix + '.face'] = ~corner_labels
    return d


if __name__ == '__main__':
    import openpnm as op
    import matplotlib.pyplot as plt

    net = fcc([3, 3, 3], 1, mode='tri', node_prefix='pore', edge_prefix='throat')
    pn = op.network.Network()
    pn.update(net)
    pn['pore.all'] = np.ones((np.shape(pn.coords)[0]), dtype=bool)
    pn['throat.all'] = np.ones((np.shape(pn.conns)[0]), dtype=bool)
    fig, ax = plt.subplots()
    op.topotools.plot_connections(pn, ax=ax)
    op.topotools.plot_coordinates(pn, ax=ax)



from .spheres_and_cylinders import spheres_and_cylinders
from .circles_and_rectangles import circles_and_rectangles
from .cones_and_cylinders import cones_and_cylinders
from .pyramids_and_cuboids import pyramids_and_cuboids
from .trapezoids_and_rectangles import trapezoids_and_rectangles
from .cubes_and_cuboids import cubes_and_cuboids
from .squares_and_rectangles import squares_and_rectangles



import logging
import numpy as np
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import Docorator
from openpnm.integrators import ScipyRK45
from openpnm.algorithms._solution import SolutionContainer


__all__ = ['TransientReactiveTransport']


docstr = Docorator()
logger = logging.getLogger(__name__)


@docstr.get_sections(base='TransientReactiveTransportSettings',
                     sections=['Parameters', 'Other Parameters'])
@docstr.dedent
class TransientReactiveTransportSettings:
    r"""

    Parameters
    ----------
    %(ReactiveTransportSettings.parameters)s

    """
    pore_volume = 'pore.volume'


class TransientReactiveTransport(ReactiveTransport):
    r"""
    A subclass of ReactiveTransport for transient simulations.

    Parameters
    ----------
    network : Network
        The Network with which this algorithm is associated.

    Notes
    -----
    Either a Network or a Project must be specified.

    """

    def __init__(self, phase, name='trans_react_?', **kwargs):
        super().__init__(phase=phase, name=name, **kwargs)
        self.settings._update(TransientReactiveTransportSettings())
        self.settings['phase'] = phase.name
        self["pore.ic"] = np.nan

    def run(self, x0, tspan, saveat=None, integrator=None):
        """
        Runs the transient algorithm and returns the solution.

        Parameters
        ----------
        x0 : ndarray or float
            Array (or scalar) containing initial condition values.
        tspan : array_like
            Tuple (or array) containing the integration time span.
        saveat : array_like or float, optional
            If an array is passed, it signifies the time points at which
            the solution is to be stored, and if a scalar is passed, it
            refers to the interval at which the solution is to be stored.
        integrator : Integrator, optional
            Integrator object which will be used to to the time stepping.
            Can be instantiated using openpnm.integrators module.

        Returns
        -------
        TransientSolution
            The solution object, which is basically a numpy array with
            the added functionality that it can be called to return the
            solution at intermediate times (i.e., those not stored in the
            solution object).

        """
        logger.info('Running TransientTransport')
        if np.isscalar(saveat):
            saveat = np.arange(*tspan, saveat)
        # FIXME: why do we forcibly add tspan[1] to saveat even if the user
        # didn't want to?
        if (saveat is not None) and (tspan[1] not in saveat):
            saveat = np.hstack((saveat, [tspan[1]]))
        integrator = ScipyRK45() if integrator is None else integrator
        # Perform pre-solve validations
        self._validate_settings()
        self._validate_topology_health()
        self._validate_linear_system()
        # Write x0 to algorithm the obj (needed by _update_iterative_props)
        self['pore.ic'] = x0 = np.ones(self.Np, dtype=float) * x0
        self._merge_inital_and_boundary_values()
        # Build RHS (dx/dt = RHS), then integrate the system of ODEs
        rhs = self._build_rhs()
        # Integrate RHS using the given solver
        soln = integrator.solve(rhs, x0, tspan, saveat)
        # Return solution as dictionary
        self.soln = SolutionContainer()
        self.soln[self.settings['quantity']] = soln

    def _run_special(self, x0):
        pass

    def _build_rhs(self):
        """
        Returns a function handle, which calculates dy/dt = rhs(y, t).

        Notes
        -----
        ``y`` is the variable that the algorithms solves for, e.g., for
        ``TransientFickianDiffusion``, it would be concentration.

        """
        def ode_func(t, y):
            # TODO: add a cache mechanism
            self.x = y
            self._update_A_and_b()
            A = self.A.tocsc()
            b = self.b
            V = self.network[self.settings["pore_volume"]]
            return (-A.dot(y) + b) / V  # much faster than A*y

        return ode_func

    def _merge_inital_and_boundary_values(self):
        x0 = self['pore.ic']
        bc_pores = ~np.isnan(self['pore.bc.value'])
        x0[bc_pores] = self['pore.bc.value'][bc_pores]
        quantity = self.settings['quantity']
        self[quantity] = x0



"""
Basic Math
==========

"""
import logging
import numpy as np
logger = logging.getLogger(__name__)


__all__ = [
    'blank',
    'clip',
    'constant',
    'difference',
    'fraction',
    'invert',
    'normalize',
    'scaled',
    'summation',
    'product',
]


def blank(target):
    """Blank model used in PNM format, acting as a placeholder"""
    pass


def invert(target, prop):
    r"""
    Inverts the given array

    Parameters
    ----------
    target : Base
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    prop : str
        Dictionary key pointing the values to invert

    """
    return 1.0/target[prop]


def difference(target, props):
    r"""
    Subtracts elements 1:N in `props` from element 0

    Parameters
    ----------
    target : OpenPNM dict
        The object to which the model is associated
    props : list
        A list of dict keys containing the values to operate on.  If the first
        element is A, and the next are B and C, then the results is A - B - C.
    """
    A = target[props[0]]
    for B in props[1:]:
        A = A - target[B]
    return A


def fraction(target, numerator, denominator):
    r"""
    Calculates the ratio between two values

    Parameters
    ----------
    target : Base
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    numerator : str
        Dictionary key pointing the numerator values
    denominator : str
        Dictionary key pointing the denominator values

    """
    x = target[numerator]
    y = target[denominator]
    return x/y


def summation(target, props=[]):
    r"""
    Sums the values in the given arrays

    Parameters
    ----------
    target : Base
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    props : list of dictionary keys
        The dictionary keys pointing the arrays whose elements should be summed

    """
    vals = np.zeros_like(target[props[0]])
    for item in props:
        vals += target[item]
    return vals


def normalize(target, prop, xmin=0, xmax=1):
    r"""
    Normalizes the given array between the supplied limits

    Parameters
    ----------
    target : Base
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    xmin : float
        Lower limit of the re-scaled data
    xmax : float
        Upper limit of the re-scaled data

    """
    vals = target[prop]
    # Scale to 0 to 1
    vals = (vals - vals.min())/(vals.max() - vals.min())
    vals = vals*(xmax - xmin) + xmin
    return vals


def clip(target, prop, xmax, xmin=0):
    r"""
    Clips the given array within the supplied limits

    Parameters
    ----------
    target : Base
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    xmin : float
        Values below this limit will be replaced with ``xmin``.
    xmax : float
        Values above this limit will be replaced with ``xmax``.

    """
    vals = np.clip(target[prop], xmin, xmax)
    return vals


def constant(target, value):
    r"""
    Places the given constant value into the target object

    Parameters
    ----------
    target : Base
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    value : scalar
        The numerical value to apply

    Returns
    -------
    value : ndarray
        Array containing constant values equal to ``value``.

    Notes
    -----
    This model is mostly useless and for testing purposes, but might be used
    to 'reset' an array back to a default value.

    """
    return value


def product(target, props):
    r"""
    Calculates the product of multiple property values

    Parameters
    ----------
    target : Base
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    props : list[str]
        The name of the arguments to be used for the product.

    Returns
    -------
    value : ndarray
        Array containing product values of ``target[props[0]]``,
        ``target[props[1]]``, etc.

    """
    value = np.ones_like(target[props[0]])
    for item in props:
        value *= target[item]
    return value


def scaled(target, prop, factor):
    r"""
    Scales an existing value by a factor.

    Useful for constricting some throat property.

    Parameters
    ----------
    target : Base
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    prop : str
        The dictionary key of the array containing the values to be scaled.
    factor : str
        The factor by which the values should be scaled.

    Returns
    -------
    value : ndarray
        Array containing ``target[prop]`` values scaled by ``factor``.

    """
    value = target[prop]*factor
    return value



import numpy as np
from numpy import pi as _pi
from openpnm.models.geometry import _geodocs

__all__ = ["sphere",
           "cube",
           "circle",
           "square",
           "effective"
           ]


@_geodocs
def sphere(
    network,
    pore_diameter='pore.diameter'
):
    r"""
    Calculate pore volume from diameter assuming a spherical pore body

    Parameters
    ----------
    %(network)s
    %(Dp)s

    Returns
    -------
    volumes : ndarray
        Numpy ndarray containing pore volume values

    """
    return 4/3*_pi*(network[pore_diameter]/2)**3


@_geodocs
def cube(
    network,
    pore_diameter='pore.diameter'
):
    r"""
    Calculate pore volume from diameter assuming a cubic pore body

    Parameters
    ----------
    %(network)s
    %(Dp)s

    Returns
    -------

    """
    return network[pore_diameter]**3


@_geodocs
def circle(
    network,
    pore_diameter='pore.diameter',
):
    r"""
    Calculate pore volume from diameter assuming a spherical pore body

    Parameters
    ----------
    %(network)s
    %(Dp)s

    Returns
    -------

    """
    return _pi/4 * network[pore_diameter]**2


@_geodocs
def square(
    network,
    pore_diameter='pore.diameter',
):
    r"""
    Calculate pore volume from diameter assuming a cubic pore body

    Parameters
    ----------
    %(network)s
    %(Dp)s

    Returns
    -------

    """
    return network[pore_diameter]**2


@_geodocs
def effective(
    network,
    pore_volume='pore.volume',
    throat_volume='throat.volume',
):
    r"""
    Calculate the effective pore volume for optional use in transient
    simulations. The effective pore volume is calculated by adding half
    the volume of all neighbouring throats to the pore volume.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    """
    cn = network['throat.conns']
    P1 = cn[:, 0]
    P2 = cn[:, 1]
    eff_vol = np.copy(network[pore_volume])
    np.add.at(eff_vol, P1, 1/2*network[throat_volume])
    np.add.at(eff_vol, P2, 1/2*network[throat_volume])
    return eff_vol



import os
import json
import pickle
import logging
import numpy as np
from pathlib import Path
from openpnm.io import _parse_filename
from openpnm.network import Network
logger = logging.getLogger(__name__)


def _validate_json(json_file):
    import jsonschema
    # Validate name of schema file
    relative_path = '../../utils/jgf_schema.pkl'
    schema_file = Path(os.path.realpath(__file__), relative_path)
    schema_file = _parse_filename(filename=schema_file, ext='pkl')

    # Load schema from pickle file
    with open(schema_file, 'rb') as file:
        jgf_schema = pickle.load(file)

    # Validate JSON agains schema
    try:
        jsonschema.validate(json_file, jgf_schema)
        return True
    except jsonschema.exceptions.ValidationError:
        return False


def network_to_jsongraph(network, filename=''):
    r"""
    Write the network to disk as a JGF file.

    Parameters
    ----------
    network : Network

    filename : str
        Desired file name, defaults to network name if not given
    """

    # Ensure output file is valid
    filename = _parse_filename(filename=filename, ext='json')

    # Ensure network contains the required properties
    try:
        required_props = {'pore.diameter', 'pore.coords', 'throat.length',
                          'throat.conns', 'throat.diameter'}
        assert required_props.issubset(network.props())
    except AssertionError:
        raise Exception('Error - network is missing one of: '
                        + str(required_props))

    # Create 'metadata' JSON object
    graph_metadata_obj = {'number_of_nodes': network.Np,
                          'number_of_links': network.Nt}

    # Create 'nodes' JSON object
    nodes_obj = [
        {
            'id': str(ps),
            'metadata': {
                'node_squared_radius': int(network['pore.diameter'][ps] / 2)**2,
                'node_coordinates': {
                    'x': int(network['pore.coords'][ps, 0]),
                    'y': int(network['pore.coords'][ps, 1]),
                    'z': int(network['pore.coords'][ps, 2])
                }
            }
        } for ps in network.Ps]

    # Create 'edges' JSON object
    edges_obj = [
        {
            'id': str(ts),
            'source': str(network['throat.conns'][ts, 0]),
            'target': str(network['throat.conns'][ts, 1]),
            'metadata': {
                'link_length': float(network['throat.length'][ts]),
                'link_squared_radius': float(network['throat.diameter'][ts] / 2)**2
            }
        } for ts in network.Ts]

    # Build 'graph' JSON object from 'metadata', 'nodes' and 'edges'
    graph_obj = {'metadata': graph_metadata_obj,
                 'nodes': nodes_obj,
                 'edges': edges_obj}

    # Build full JSON object
    json_obj = {'graph': graph_obj}

    # Load and validate input JSON
    with open(filename, 'w') as file:
        json.dump(json_obj, file, indent=2)


def network_from_jsongraph(filename):
    r"""
    Loads the JGF file onto the given project.

    Parameters
    ----------
    filename : str
        The name of the file containing the data to import.  The formatting
        of this file is outlined below.

    Returns
    -------
    network : dict
        An OpenPNM Network dictionary

    """

    # Ensure input file is valid
    filename = _parse_filename(filename=filename, ext='json')

    # Load and validate input JSON
    with open(filename, 'r') as file:
        json_file = json.load(file)
        if not _validate_json(json_file):
            raise Exception('File is not in the JSON Graph Format')

    # Extract graph metadata from JSON
    number_of_nodes = json_file['graph']['metadata']['number_of_nodes']
    number_of_links = json_file['graph']['metadata']['number_of_links']

    # Extract node properties from JSON
    nodes = sorted(json_file['graph']['nodes'], key=lambda node: int(node['id']))
    x = np.array([node['metadata']['node_coordinates']['x'] for node in nodes])
    y = np.array([node['metadata']['node_coordinates']['y'] for node in nodes])
    z = np.array([node['metadata']['node_coordinates']['z'] for node in nodes])

    # Extract link properties from JSON
    edges = sorted(json_file['graph']['edges'], key=lambda edge: int(edge['id']))
    source = np.array([int(edge['source']) for edge in edges])
    target = np.array([int(edge['target']) for edge in edges])
    link_length = np.array([edge['metadata']['link_length'] for edge in edges])
    link_squared_radius = np.array(
        [edge['metadata']['link_squared_radius'] for edge in edges])

    # Generate network object
    network = Network()
    network['pore.all'] = np.ones([number_of_nodes, ], dtype=bool)
    network['throat.all'] = np.ones([number_of_links, ], dtype=bool)

    # Define primitive throat properties
    network['throat.length'] = link_length
    network['throat.conns'] = np.column_stack([source, target])
    network['throat.diameter'] = 2.0 * np.sqrt(link_squared_radius)

    # Define primitive pore properties
    network['pore.index'] = np.arange(number_of_nodes)
    network['pore.coords'] = np.column_stack([x, y, z])
    network['pore.diameter'] = np.zeros(number_of_nodes)

    return network



r"""
Topology
--------
Pore-scale models related to topology of the network.

"""
from numpy.linalg import norm
import numpy as np
import scipy.spatial as sptl


__all__ = [  # Keep this alphabetical for easier inspection of what's imported
    'coordination_number',
    'distance_to_furthest_neighbor',
    'distance_to_nearest_neighbor',
    'distance_to_nearest_pore',
    'gabriel_edges',
    'pore_to_pore_distance',
]


def gabriel_edges(network):
    r"""
    Find throats which make a Gabriel subgraph

    Returns
    -------
    throats : ndarray
        An ndarray of boolean values with ``True`` indicating that a throat
        satisfies the conditions of Gabriel graph, meaning that a circle (or
        sphere) can be drawn between its two connected pores that does not
        contain any other pores.

    Notes
    -----
    Technically this should only be used on a Delaunay network, but it will
    work on any graph. By deleting all throats that are *not* identified by
    this fuction one would obtain the Gabriel graph [1].

    References
    ----------
    [1] `Wikipedia <https://en.wikipedia.org/wiki/Gabriel_graph>`_

    """
    dn = network
    # Find centroid or midpoint of each edge in conns
    c = dn['pore.coords'][dn['throat.conns']]
    m = (c[:, 0, :] + c[:, 1, :])/2
    # Find the radius the sphere between each pair of nodes
    r = np.sqrt(np.sum((c[:, 0, :] - c[:, 1, :])**2, axis=1))/2
    # Use the kd-tree function in Scipy's spatial module
    tree = sptl.cKDTree(dn['pore.coords'])
    # Find the nearest point for each midpoint
    n = tree.query(x=m, k=1)[0]
    # If nearest point to m is at distance r, then the edge is a Gabriel edge
    g = n >= r*(0.999)  # This factor avoids precision errors in the distances
    return g


def coordination_number(network):
    r"""
    Find the number of neighbors for each pore
    """
    N = network.num_neighbors(pores=network.Ps, flatten=False)
    return N


def pore_to_pore_distance(network):
    r"""
    Find the center to center distance between each pair of pores
    """
    cn = network['throat.conns']
    C1 = network['pore.coords'][cn[:, 0]]
    C2 = network['pore.coords'][cn[:, 1]]
    values = norm(C1 - C2, axis=1)
    return values


def distance_to_nearest_neighbor(network):
    r"""
    Find the distance between each pore and its closest topological neighbor
    """
    cn = network['throat.conns']
    C1 = network['pore.coords'][cn[:, 0]]
    C2 = network['pore.coords'][cn[:, 1]]
    D = norm(C1 - C2, axis=1)
    im = network.create_incidence_matrix()
    values = np.ones((network.Np, ))*np.inf
    np.minimum.at(values, im.row, D[im.col])
    return np.array(values)


def distance_to_furthest_neighbor(network):
    r"""
    Find the distance between each pore and its furthest topological neighbor
    """
    throats = network.throats(network.name)
    cn = network['throat.conns'][throats]
    C1 = network['pore.coords'][cn[:, 0]]
    C2 = network['pore.coords'][cn[:, 1]]
    D = norm(C1 - C2, axis=1)
    im = network.create_incidence_matrix()
    values = np.zeros((network.Np, ))
    np.maximum.at(values, im.row, D[im.col])
    return np.array(values)


def distance_to_nearest_pore(network):
    r"""
    Find distance to and index of nearest pore even if not topologically
    connected
    """
    import scipy.spatial as sptl
    coords = network.coords
    tree = sptl.KDTree(coords)
    ds, ids = tree.query(coords, k=2)
    values = ds[:, 1]
    return values



import logging
import numpy as np
import scipy.sparse as sprs
from scipy.sparse import csgraph
from openpnm.utils import PrintableDict, Workspace
from openpnm._skgraph import simulations
from openpnm._skgraph import queries
from collections import namedtuple


logger = logging.getLogger(__name__)
ws = Workspace()
__all__ = [
    'ispercolating',
    'find_isolated_clusters',
    'trim_disconnected_clusters',
    'site_percolation',
    'bond_percolation',
    'find_clusters',
    'find_path',
]


def find_path(network, pore_pairs, weights=None):
    return queries.find_path(network=network, pairs=pore_pairs, weights=weights)


find_path.__doc__ = queries.find_path.__doc__


def ispercolating(network, inlets, outlets):
    if np.array(inlets).dtype == bool:
        inlets = np.where(inlets)[0]
    if np.array(outlets).dtype == bool:
        outlets = np.where(outlets)[0]
    flag = simulations.ispercolating(
        conns=network.conns,
        occupied=np.ones(network.Nt, dtype=bool),
        inlets=inlets,
        outlets=outlets
    )
    return flag


ispercolating.__doc__ = simulations.ispercolating.__doc__


def site_percolation(network, occupied_sites):
    return simulations.site_percolation(network.conns, occupied_sites)


site_percolation.__doc__ = simulations.site_percolation.__doc__


def bond_percolation(network, occupied_bonds):
    return simulations.bond_percolation(network.conns, occupied_bonds)


bond_percolation.__doc__ = simulations.bond_percolation.__doc__


def trim_disconnected_clusters(**kwargs):
    return simulations.trim_disconnected_clusters(**kwargs)


trim_disconnected_clusters.__doc__ = \
    simulations.trim_disconnected_clusters.__doc__


def find_isolated_clusters(network, mask, inlets):
    r"""
    Identifies pores and throats that are invaded but not connected to the inlets

    Parameters
    ----------
    network : dict
        The OpenPNM Network
    mask : ndarray
        A boolean mask of either Nt or Np length with ``True`` values
        indicating invaded bonds or sites. If this array is Nt-long then
        then bond percolation is used to identify clusters, whereas site
        percolation is used if it is Np-long.
    inlets : ndarray
        A array containing indices of the pores which define the inlets. Any
        clusters not connected to these sites are considered isolated.

    Returns
    -------
    sites : ndarray
        An ndarray containing the indices of invaded pores which are not
        connected to the given ``inlets``.
    """
    labels = find_clusters(network=network, mask=mask)
    isolated = np.in1d(labels.pore_labels, labels.pore_labels[inlets], invert=True)
    isolated = np.where(isolated)[0]
    return isolated


def find_clusters(network, mask=[]):
    r"""
    Identify connected clusters of pores and throats in the network.

    Either site and bond percolation can be considered, see description of
    ``mask`` argument for details.

    Parameters
    ----------
    network : Network
        The network
    mask : array_like, boolean
        A list of open bonds or sites (throats or pores). If the mask is
        Np-long, then the method will perform a site percolation to identify
        clusters, and if the mask is Nt-long bond percolation will be
        performed.

    Returns
    -------
    p_labels, t_labels : tuple of ndarrays
        A tuple containing an Np-long array of pore cluster labels, and an
        Nt-long array of throat cluster labels. The label numbers correspond
        such that pores and throats with the same label are part of the same
        cluster. Uninvaded locations are set to -1.

    """
    # Parse the input arguments
    mask = np.array(mask, ndmin=1)
    if mask.dtype != bool:
        raise Exception('Mask must be a boolean array of Np or Nt length')
    # If pore mask was given perform site percolation
    if np.size(mask) == network.Np:
        (p_clusters, t_clusters) = \
            simulations.site_percolation(network.conns, mask)
    # If pore mask was given perform bond percolation
    elif np.size(mask) == network.Nt:
        (p_clusters, t_clusters) = \
            simulations.bond_percolation(network.conns, mask)
    else:
        raise Exception('Mask received was neither Nt nor Np long')
    result = namedtuple('result', ('pore_labels', 'throat_labels'))
    result.pore_labels = p_clusters
    result.throat_labels = t_clusters
    return result



from numpy import pi as _pi
from openpnm.models.geometry import _geodocs


__all__ = [
    "cone",
    "sphere",
    "cube",
    "circle",
    "square"
]


@_geodocs
def sphere(
    network,
    pore_diameter='pore.diameter'
):
    r"""
    Calculate cross-sectional area assuming the pore body is a sphere

    Parameters
    ----------
    %(network)s
    %(Dp)s

    Returns
    -------
    areas : ndarray
        A numpy ndarry containing pore cross-sectional area values

    """
    D = network[pore_diameter]
    return _pi/4 * D**2


@_geodocs
def cone(
    network,
    pore_diameter='pore.diameter'
):
    r"""
    Calculate cross-sectional area assuming the pore body is a cone

    Parameters
    ----------
    %(network)s
    %(Dp)s

    Returns
    -------

    """
    D = network[pore_diameter]
    return _pi / 4 * D**2


@_geodocs
def cube(
    network,
    pore_diameter='pore.diameter'
):
    r"""
    Calculate cross-sectional area assuming the pore body is a cube

    Parameters
    ----------
    %(network)s
    %(Dp)s

    Returns
    -------

    """
    D = network[pore_diameter]
    return D**2


@_geodocs
def circle(
    network,
    pore_diameter='pore.diameter'
):
    r"""
    Calculate cross-sectional area assuming the pore body is a circle

    Parameters
    ----------
    %(network)s
    %(Dp)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    return network[pore_diameter]


@_geodocs
def square(
    network,
    pore_diameter='pore.diameter'
):
    r"""
    Calculate cross-sectional area assuming the pore body is a square

    Parameters
    ----------
    %(network)s
    %(Dp)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    return network[pore_diameter]



import logging
import numpy as np
from openpnm.core import ParserMixin, LabelMixin, Base2
from openpnm.utils import Docorator


__all__ = ['Algorithm']


logger = logging.getLogger(__name__)
docstr = Docorator()


@docstr.get_sections(base='AlgorithmSettings', sections=docstr.all_sections)
@docstr.dedent
class AlgorithmSettings:
    r"""

    Parameters
    ----------
    %(BaseSettings.parameters)s

    """


@docstr.get_sections(base='Algorithm', sections=['Parameters'])
@docstr.dedent
class Algorithm(ParserMixin, LabelMixin, Base2):
    r"""
    Generic class to define the foundation of Algorithms

    Parameters
    ----------
    %(Base.parameters)s

    """

    def __init__(self, network, name='alg_?', **kwargs):
        super().__init__(network=network, name=name, **kwargs)
        self.settings._update(AlgorithmSettings())
        self['pore.all'] = np.ones([network.Np, ], dtype=bool)
        self['throat.all'] = np.ones([network.Nt, ], dtype=bool)

    # @functools.cached_property
    @property
    def iterative_props(self):
        r"""
        Finds and returns properties that need to be iterated while
        running the algorithm.
        """
        import networkx as nx
        phase = self.project[self.settings.phase]
        # Generate global dependency graph
        dg = nx.compose_all([x.models.dependency_graph(deep=True)
                             for x in [phase]])
        variable_props = self.settings["variable_props"].copy()
        variable_props.add(self.settings["quantity"])
        base = list(variable_props)
        # Find all props downstream that depend on base props
        dg = nx.DiGraph(nx.edge_dfs(dg, source=base))
        if len(dg.nodes) == 0:
            return []
        iterative_props = list(nx.dag.lexicographical_topological_sort(dg))
        # "variable_props" should be in the returned list but not "quantity"
        if self.settings.quantity in iterative_props:
            iterative_props.remove(self.settings["quantity"])
        return iterative_props

    def _update_iterative_props(self, iterative_props=None):
        """
        Regenerates phase, geometries, and physics objects using the
        current value of ``quantity``.

        Notes
        -----
        The algorithm directly writes the value of 'quantity' into the
        phase, which is against one of the OpenPNM rules of objects not
        being able to write into each other.

        """
        if iterative_props is None:
            iterative_props = self.iterative_props
        if not iterative_props:
            return
        # Fetch objects associated with the algorithm
        phase = self.project[self.settings.phase]
        # Update 'quantity' on phase with the most recent value
        quantity = self.settings['quantity']
        phase[quantity] = self.x
        # Regenerate all associated objects
        phase.regenerate_models(propnames=iterative_props)

    def clear_BCs(self, bctype=[]):
        r"""
        Clear all BCs of the given type(s)

        Parameters
        ----------
        bctype : str or list(str)
            The specific type of boundary condition to clear. If not provided
            all types will be cleared.
        """
        self.set_BC(pores=None, bctype=bctype, mode='remove')

    def set_BC(self, pores=None, bctype=[], bcvalues=[], mode='add'):
        r"""
        The main method for setting and adjusting boundary conditions.

        This method is called by other more convenient wrapper functions like
        ``set_value_BC``.

        Parameters
        ----------
        pores : array_like
            The pores where the boundary conditions should be applied. If
            ``None`` is given then *all* pores are assumed.  This is useful
            when ``mode='remove'``.
        bctype : str
            Specifies the type or the name of boundary condition to apply. This
            can be anything, but normal options are 'rate' and 'value'. If
            an empty list is provided, then all bc types will be assumed. This
            is useful for clearing all bcs if ``mode='remove'`` and ``pores=
            None``.
        bcvalues : int or array_like
            The boundary value to apply, such as concentration or rate.
            If a single value is given, it's assumed to apply to all
            locations. Different values can be applied to all pores in the
            form of an array of the same length as ``pores``. Note that using
            ``mode='add'`` and ``values=np.nan`` is equivalent to removing
            bcs from the given ``pores``.
        mode : str or list of str, optional
            Controls how the boundary conditions are applied. Options are:

            ============ =====================================================
            mode         meaning
            ============ =====================================================
            'add'        (default) Adds the supplied boundary conditions to
                         the given locations. Raises an exception if values
                         of any type already exist in the given locations.
            'overwrite'  Adds supplied boundary conditions to the given
                         locations, including overwriting conditions of the
                         given type or any other type that may be present in
                         the given locations.
            'remove'     Removes boundary conditions of the specified type
                         from the specified locations. If ``bctype`` is not
                         specified then *all* types are removed. If no
                         locations are given then values are remvoed from
                         *all* locations.
            ============ =====================================================

            If a list of strings is provided, then each mode in the list is
            handled in order, so that ``['remove', 'add']`` will give the same
            results add ``'overwrite'``.

        Notes
        -----
        It is not possible to have multiple boundary conditions for a
        specified location in one algorithm.

        """
        # If a list of modes was given, handle them each in order
        if not isinstance(mode, str):
            for item in mode:
                self.set_BC(pores=pores, bctype=bctype,
                            bcvalues=bcvalues, mode=item)
            return
        # If a list of bctypes was given, handle them each in order
        if len(bctype) == 0:
            bctype = self['pore.bc'].keys()
        if not isinstance(bctype, str):
            for item in bctype:
                self.set_BC(pores=pores, bctype=item,
                            bcvalues=bcvalues, mode=mode)
            return

        # Begin method
        bc_types = list(self['pore.bc'].keys())
        other_types = np.setdiff1d(bc_types, bctype).tolist()

        mode = self._parse_mode(
            mode,
            allowed=['overwrite', 'add', 'remove'],
            single=True)

        # Infer the value that indicates "no bc" based on array dtype
        no_bc = get_no_bc(self[f'pore.bc.{bctype}'])

        if pores is None:
            pores = self.Ps
        pores = self._parse_indices(pores)

        # Deal with size of the given bcvalues
        values = np.array(bcvalues)
        if values.size == 1:
            values = np.ones_like(pores, dtype=values.dtype)*values
        # Ensure values and pores are the same size
        if values.size > 1 and values.size != pores.size:
            raise Exception('The number of values must match the number of locations')

        # Finally adjust the BCs according to mode
        if mode == 'add':
            mask = np.ones_like(pores, dtype=bool)  # Indices of pores to keep
            for item in self['pore.bc'].keys():  # Remove pores that are taken
                mask[isfinite(self[f'pore.bc.{item}'][pores])] = False
            if not np.all(mask):  # Raise exception if some conflicts found
                msg = "Some of the given locations already have BCs, " \
                    + "either use mode='remove' first or " \
                    + "use mode='overwrite' instead"
                raise Exception(msg)
            self[f"pore.bc.{bctype}"][pores[mask]] = values[mask]
        elif mode == 'overwrite':
            # Put given values in specified BC, sort out conflicts below
            self[f"pore.bc.{bctype}"][pores] = values
            # Collect indices that are present for other BCs for removal
            mask = np.ones_like(pores, dtype=bool)
            for item in other_types:
                self[f"pore.bc.{item}"][pores] = get_no_bc(self[f"pore.bc.{item}"])
                # Make a note of any BCs values of other types
                mask[isfinite(self[f'pore.bc.{item}'][pores])] = False
            if not np.all(mask):  # Warn that other values were overwritten
                msg = 'Some of the given locations already have BCs of ' \
                    + 'another type, these will be overwritten'
                logger.warning(msg)
        elif mode == 'remove':
            self[f"pore.bc.{bctype}"][pores] = no_bc


def get_no_bc(arr):
    no_bc = np.nan if arr.dtype in (float, int) else False
    return no_bc


def isfinite(arr, inf=None):
    if arr.dtype in (bool, ):
        results = arr == True
    elif arr.dtype in (float, ):
        results = ~np.isnan(arr)
    else:
        results = arr != inf
    return results



import logging
import sys
import numpy as np
import scipy as sp
import scipy.sparse
from openpnm.solvers import IterativeSolver
logger = logging.getLogger(__name__)
try:
    import petsc4py
    # Next line must be before importing PETSc
    petsc4py.init(sys.argv)
    from petsc4py import PETSc
except ModuleNotFoundError:
    pass


__all__ = ['PETScLinearSolver']


class PETScLinearSolver(IterativeSolver):
    r"""
    Solves the sparse linear system Ax = b using petsc solvers.

    Notes
    -----
    Parallel computing is supported and matrix partitioning over the
    available cores is automatically handled by running:

    .. code::

        $ mpirun -np num_cores python script.py

    where ``num_cores`` must be substituted with the number of cores.

    """
    def _create_solver(self):
        r"""
        This method creates the petsc sparse linear solver.
        """
        # https://petsc.org/release/docs/manualpages/KSP/KSPType.html
        iterative = [
            'richardson', 'chebyshev', 'cg', 'groppcg', 'pipecg', 'pipecgrr',
            'cgne', 'nash', 'stcg', 'gltr', 'fcg', 'pipefcg', 'gmres',
            'pipefgmres', 'fgmres', 'lgmres', 'dgmres', 'pgmres', 'tcqmr',
            'bcgs', 'ibcgs', 'fbcgs', 'fbcgsr', 'bcgsl', 'pipebcgs', 'cgs',
            'tfqmr', 'cr', 'pipecr', 'lsqr', 'preonly', 'qcg', 'bicg',
            'minres', 'symmlq', 'lcd', 'python', 'gcr', 'pipegcr', 'tsirm',
            'cgls', 'fetidp']
        # https://petsc.org/release/docs/manualpages/PC/PCType.html
        preconditioners = [
            'none', 'jacobi', 'sor', 'lu', 'shell', 'bjacobi', 'mg',
            'eisenstat', 'ilu', 'icc', 'asm', 'gasm', 'ksp', 'composite',
            'redundant', 'spai', 'nn', 'cholesky', 'pbjacobi', 'mat', 'hypre',
            'parms', 'fieldsplit', 'tfs', 'ml', 'galerkin', 'exotic', 'cp',
            'bfbt', 'lsc', 'python', 'pfmg', 'syspfmg', 'redistribute', 'svd',
            'gamg', 'sacusp', 'sacusppoly', 'bicgstabcusp', 'ainvcusp',
            'chowiluviennacl', 'rowscalingviennacl', 'saviennacl', 'bddc',
            'kaczmarz', 'telescope']
        direct_lu = ['mumps', 'superlu_dist', 'umfpack', 'klu']
        direct_cholesky = ['mumps', 'cholmod']
        valid_solvers = iterative + direct_lu + direct_cholesky

        solver = self.solver_type
        preconditioner = self.preconditioner

        if solver not in valid_solvers:
            raise Exception(f"{solver} solver not availabe, choose another solver")
        if preconditioner not in preconditioners:
            raise Exception(f"{preconditioner} not found, choose another preconditioner")

        self.ksp = PETSc.KSP()
        self.ksp.create(PETSc.COMM_WORLD)

        if solver in direct_lu:
            self.ksp.getPC().setType('lu')
            self.ksp.getPC().setFactorSolverType(solver)
            self.ksp.setType('preonly')
        elif solver in direct_cholesky:
            self.ksp.getPC().setType('cholesky')
            self.ksp.getPC().setFactorSolverType(solver)
            self.ksp.setType('preonly')
        elif solver in preconditioners:
            self.ksp.getPC().setType(solver)
            self.ksp.setType('preonly')
        elif solver in iterative:
            self.ksp.getPC().setType(preconditioner)
            self.ksp.setType(solver)

    def _set_tolerances(self, atol=None, rtol=None, maxiter=None):
        r"""
        Set absolute and relative tolerances, and maximum number of iterations.
        """
        atol = self.atol if atol is None else atol
        rtol = self.rtol if rtol is None else rtol
        maxiter = self.maxiter if maxiter is None else maxiter
        # BUG: PETSc misses rtol requirement by ~10-20X -> Report to petsc4py
        self.ksp.setTolerances(atol=None, rtol=rtol/50, max_it=maxiter)

    def _assemble_A(self):
        r"""
        This method creates the petsc sparse coefficients matrix from the
        OpenPNM scipy one. The method also equally decomposes the matrix at
        certain rows into different blocks (each block contains all the
        columns) and distributes them over the pre-assigned cores for parallel
        computing. The method can be used in serial.
        """

        # Create a petsc sparse matrix
        self.petsc_A = PETSc.Mat()
        self.petsc_A.create(PETSc.COMM_WORLD)
        self.petsc_A.setSizes([self.m, self.n])
        self.petsc_A.setType('aij')  # sparse
        self.petsc_A.setUp()

        # Loop over owned block of rows on this processor
        # and insert entry values (for parallel computing).
        self.Istart, self.Iend = self.petsc_A.getOwnershipRange()

        # Assign values to the coefficients matrix from the scipy sparse csr
        size_tmp = self.A.shape
        # Row indices
        csr1 = self.A.indptr[self.Istart:self.Iend+1] - self.A.indptr[self.Istart]
        ind1 = self.A.indptr[self.Istart]
        ind2 = self.A.indptr[self.Iend]
        csr2 = self.A.indices[ind1:ind2]  # column indices
        csr3 = self.A.data[ind1:ind2]  # data
        self.petsc_A = PETSc.Mat().createAIJ(size=size_tmp,
                                             csr=(csr1, csr2, csr3))
        # Communicate off-processor values and setup internal data structures
        # for performing parallel operations
        self.petsc_A.assemblyBegin()
        self.petsc_A.assemblyEnd()

    def _assemble_b_and_x(self):
        r"""
        Initialize the solution vector (self.petsc_x), which is a dense
        matrix (1D vector) and defines the rhs vector (self.petsc_b) from
        the existing data.
        """
        # Get vector(s) compatible with the matrix (same parallel layout)
        # passing same communicator as the A matrix

        # Global solution vector (all the local solutions will return to it)
        self.petsc_s = PETSc.Vec()
        self.petsc_s.create(PETSc.COMM_WORLD)
        self.petsc_s.setSizes(self.m)
        self.petsc_s.setFromOptions()

        self.Istart, self.Iend = self.petsc_s.getOwnershipRange()

        self.petsc_x = (self.petsc_s).duplicate()
        self.petsc_b = (self.petsc_s).duplicate()

        # Set the solution vector to zeros or the given initial guess (if any)
        PETSc.Vec.setArray(self.petsc_x, self.x0[self.Istart: self.Iend])

        # Define the petsc rhs vector from the numpy one
        PETSc.Vec.setArray(self.petsc_b, self.b[self.Istart: self.Iend])

    def solve(self, A, b, x0=None, solver_type='cg', precondioner='jacobi',
              maxiter=None, atol=None, rtol=None):
        r"""
        Solves and returns the solution to the linear system, Ax = b.

        This method converts the solution vector from a PETSc.Vec
        instance to a numpy array, and finally destroys all the PETSc
        objects to free memory.

        Parameters
        ----------
        A : csr_matrix
            Coefficients matrix in Ax = b
        b : ndarray
            Right-hand-side vector in Ax = b
        solver_type : str, optional
            Default is the iterative solver 'cg' based on the
            Conjugate Gradient method.
        preconditioner: str, optional
            Default is the 'jacobi' preconditioner, i.e., diagonal
            scaling preconditioning. The preconditioner is used with
            iterative solvers. When a direct solver is used, this
            parameter is ignored.
        factorization_type : str, optional
            The factorization type used with the direct solver.
            Default is 'lu'. This parameter is ignored when an
            iterative solver is used.

        Returns
        -------
        ndarray
            The solution to Ax = b

        Notes
        -----
        Certain combinations of iterative solvers and precondioners
        or direct solvers and factorization types are not supported.
        The summary table of the different possibilities
        can be found
        `here <https://petsc.org/main/overview/linear_solve_table>`_

        """
        self.b = b
        self.A = sp.sparse.csr_matrix(A)
        self.m, self.n = self.A.shape
        self.x0 = np.zeros_like(self.b) if x0 is None else x0

        self.solver_type = solver_type
        self.preconditioner = precondioner

        self.atol = self._get_atol(self.b)
        self.rtol = self._get_rtol(self.x0)

        self._assemble_b_and_x()
        self._assemble_A()

        self._create_solver()
        self._set_tolerances(atol=atol, rtol=rtol, maxiter=maxiter)
        self.ksp.setOperators(self.petsc_A)
        self.ksp.setFromOptions()

        # Solve the linear system
        self.ksp.solve(self.petsc_b, self.petsc_x)

        # Gather the solution to all processors
        gather_to_0, self.petsc_s = PETSc.Scatter().toAll(self.petsc_x)
        gather_to_0.scatter(self.petsc_x, self.petsc_s,
                            PETSc.InsertMode.INSERT, PETSc.ScatterMode.FORWARD)

        # Convert solution vector from PETSc.Vec instance to a numpy array
        self.solution = PETSc.Vec.getArray(self.petsc_s)

        # Destroy petsc solver, coefficients matrix, rhs, and solution vectors
        PETSc.KSP.destroy(self.ksp)
        PETSc.Mat.destroy(self.petsc_A)
        PETSc.Vec.destroy(self.petsc_b)
        PETSc.Vec.destroy(self.petsc_x)
        PETSc.Vec.destroy(self.petsc_s)

        # FIXME: fetch exit_code somehow from petsc
        exit_code = 0

        return self.solution, exit_code



from transforms3d import _gohlketransforms as tr
from openpnm.models.geometry import _geodocs


__all__ = ["pore_to_pore"]


@_geodocs
def pore_to_pore(network):
    r"""
    Calculates throat vector as straight path between connected pores.

    Parameters
    ----------
    %(network)s

    Returns
    -------
    unit_vec : ndarray
        A [Nt-by-3] numpy ndarray containing pore-to-pore unit vectors

    Notes
    -----
    There is an important impicit assumption here: the positive direction is
    taken as the direction from the pore with the lower index to the higher.
    This corresponds to the pores in the 1st and 2nd columns of the
    'throat.conns' array as stored on the network.

    """
    conns = network['throat.conns']
    P1 = conns[:, 0]
    P2 = conns[:, 1]
    coords = network['pore.coords']
    vec = coords[P2] - coords[P1]
    unit_vec = tr.unit_vector(vec, axis=1)
    return unit_vec



from openpnm.models import misc as _misc


__all__ = ["random",
           "from_neighbor_pores"]


def random(network, seed=None, num_range=[0, 1]):
    return _misc.random(network, element='throat', seed=seed,
                        num_range=num_range)


random.__doc__ = _misc.random.__doc__


def from_neighbor_pores(network, prop='pore.seed', mode='min'):
    return _misc.from_neighbor_pores(network, prop=prop,
                                     mode=mode)


random.__doc__ = _misc.from_neighbor_pores.__doc__



import numpy as np
from openpnm.utils import HealthDict


__all__ = [
    'check_data_health',
    'check_network_health',
]


def check_data_health(obj):
    r"""
    Checks the health of pore and throat data arrays.

    Parameters
    ----------
    obj : Base
        A handle of the object to be checked

    Returns
    -------
    health : dict
        Returns a HealthDict object which is a basic dictionary with an
        added ``health`` attribute that is ``True`` is all entries in the
        dict are deemed healthy (empty lists), or ``False`` otherwise.

    """
    health = HealthDict()
    for item in obj.props():
        health[item] = []
        if obj[item].dtype == 'O':
            health[item] = 'No checks on object'
        elif np.sum(np.isnan(obj[item])) > 0:
            health[item] = 'Has NaNs'
        elif np.shape(obj[item])[0] != obj._count(item.split('.', 1)[0]):
            health[item] = 'Wrong Length'
    return health


def check_network_health(network):
    r"""
    This method checks the topological health of the network

    The following aspects are checked for:

        (1) Isolated pores
        (2) Disconnected clusters of pores
        (3) Duplicate throats
        (4) Headless throats
        (5) Bidirectional throats

    Returns
    -------
    health : dict
        A dictionary containing the offending pores or throat numbers
        under each named key.

    Notes
    -----
    It also returns a list of which pores and throats should be trimmed
    from the network to restore health.  This list is a suggestion only,
    and is based on keeping the largest cluster and trimming the others.

    - Does not yet check for duplicate pores
    - Does not yet suggest which throats to remove
    - This is just a 'check' and does not 'fix' the problems it finds

    """
    import openpnm.models.network as mods

    health = HealthDict()

    # Check for headless throats
    headless = mods.headless_throats(network)
    health['headless_throats'] = np.where(headless)[0].tolist()

    # Check for throats that loop back onto the same pore
    looped = mods.looped_throats(network)
    health['looped_throats'] = np.where(looped)[0].tolist()

    # Check for individual isolated pores
    isolated = mods.isolated_pores(network)
    health['isolated_pores'] = np.where(isolated)[0].tolist()

    # Check for separated clusters of pores
    size = mods.cluster_size(network)
    mx = np.max(size)
    health['disconnected_pores'] = np.where(size < mx)[0].tolist()

    # Check for duplicate throats
    dupes = mods.duplicate_throats(network)
    health['duplicate_throats'] = np.where(dupes)[0].tolist()

    # Check for bidirectional throats
    bidir = mods.bidirectional_throats(network)
    health['bidirectional_throats'] = np.where(bidir)[0].tolist()

    return health



import numpy as np

from openpnm.models.phase import _phasedocs

__all__ = [
    "water_correlation",
    "liquid_pure_bb",
    "liquid_mixture_wsd",
]


@_phasedocs
def water_correlation(
    phase,
    T='pore.temperature',
    salinity='pore.salinity'
):
    r"""
    Calculates surface tension of pure water or seawater at atmospheric
    pressure

    This model uses Eq. (28) given by Sharqawy et al. [1]. Values at
    temperature higher than the normal boiling temperature are calculated at
    the saturation pressure.

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(salinity)s

    Returns
    -------
    value : ndarray
        A numpy ndarray containing surface tension of seawater in [N/m]

    Notes
    -----
    T must be in K, and S in g of salt per kg of phase, or ppt (parts per
    thousand). The correlation is valid for 273 < T < 313 K and
    0 < S < 40 g/kg within 0.2 percent accuracy.

    References
    ----------
    [1] Sharqawy M. H., Lienhard J. H., and Zubair, S. M., Desalination and
    Water Treatment, 2010.

    """
    T = phase[T]
    if salinity in phase.keys():
        S = phase[salinity]
    else:
        S = 0
    sigma_w = 0.2358*((1-(T/647.096))**1.256)*(1-0.625*(1-(T/647.096)))
    a1 = 2.2637334337E-04
    a2 = 9.4579521377E-03
    a3 = 3.3104954843E-02
    TC = T-273.15
    sigma_sw = sigma_w*(1+(a1*TC+a2)*np.log(1+a3*S))
    value = sigma_sw
    return value


@_phasedocs
def liquid_pure_bb(
    phase,
    T='pore.temperature',
    Tc='param.critical_temperature',
    Tb='param.boiling_temperature',
    Pc='param.critical_pressure',
):
    r"""
    Computes the surface tension of a pure liquid in its own vapor using the
    correlation in [1]

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(Tc)s
    %(Tb)s
    %(Pc)s

    Returns
    -------
    value : ndarray
        A numpy ndarray containing surface tension values scaled to the
        temperature [N/m]

    References
    ----------
    [1] Brock and Bird

    """
    T = phase[T]
    Tc = phase[Tc]
    Tr = T/Tc
    Tb = phase[Tb]
    Tbr = Tb/Tc
    Pc = phase[Pc]/100000
    Q = 0.1196*(1 + Tbr*np.log(Pc/1.01325)/(1-Tbr))-0.279
    sigma = 1e-3 * Q * (Pc**(2/3)) * (Tc**(1/3)) * ((1-Tr)**(11/9))
    return sigma


@_phasedocs
def liquid_mixture_wsd(
    phase,
    sigmas='pore.surface_tension.*',
    rhos='pore.density.*',
    MWs='param.molecular_weight.*',
):
    r"""
    Computes the surface tension of a liqiud mixture with its own vapor using
    the correlation in [1]

    Parameters
    ----------
    %(phase)s
    %(rhos)s
    %(MWs)s
    %(sigmas)s

    Returns
    -------
    sigma : ndarray
        A numpy ndarray containing surface tension values

    References
    ----------
    [1] Winterfeld, Scriven, and Davis

    """
    sigmas = phase.get_comp_vals(sigmas)
    xs = phase['pore.mole_fraction']
    rhos = phase.get_comp_vals(rhos)  # kg/m3
    MWs = phase.get_comp_vals(MWs)  # g/mol
    rhoms = {k: rhos[k]/(MWs[k]/1000) for k in xs.keys()}  # mol/m3
    rhom_mix = np.vstack([rhoms[k]*xs[k] for k in xs.keys()]).sum(axis=0)
    sigma = 0.0
    for i, ki in enumerate(xs.keys()):
        for j, kj in enumerate(xs.keys()):
            num = xs[ki]*xs[kj]*(rhom_mix**2)*(sigmas[ki]*sigmas[kj])**0.5
            denom = rhoms[ki]*rhoms[kj]
            sigma += num/denom
    return sigma



import numpy as np
from openpnm.models.phase import _phasedocs


__all__ = [
    'salinity',
    'mixing_rule',
    'mole_summation',
    'from_component',
    'mole_to_mass_fraction',
]


@_phasedocs
def salinity(
    phase,
    T='pore.temperature',
    conc='pore.concentration',
):
    r"""
    Calculates the salinity in g salt per kg of solution from concentration

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(conc)s

    Returns
    -------
    salinity : ndarray
        The salinity in g of solute per kg of solution.

    Notes
    -----
    This model is useful for converting known concentration values (e.g.
    calculated by a transport algorithm) into salinity values, which can then
    be used for finding other physical properties of water which are available
    as a function of salinity.

    The salinity correlations are valid for salinity up to 160 g/kg, which
    corresponds to a concentration of 0.05 mol/L (assuming NaCl is the only
    solute)

    """
    C = phase[conc]
    T = phase[T]
    a = 8.73220929e+00
    b = 6.00389629e+01
    c = -1.19083743e-01
    d = -1.77796042e+00
    e = 3.26987130e-04
    f = -1.09636011e-01
    g = -1.83933426e-07
    S = a + b*C + c*T + d*C**2 + e*T**2 + f*C**3 + g*T**3
    return S


@_phasedocs
def mixing_rule(
    phase,
    prop,
    mode='logarithmic',
    power=1,
):
    r"""
    Computes the property of a mixture using the specified mixing rule

    Parameters
    ----------
    %(phase)s
    prop : str
        The dictionary key containing the property of interest on each
        component
    mode : str
        The mixing rule to to use. Options are:

        ============== ========================================================
        mode
        ============== ========================================================
        'logarithmic'  (default) Uses the natural logarithm of the property as:
                       :math:`ln(z) = \Sigma (x_i \cdot ln(\z_i))`
        'linear'       Basic mole fraction weighting of the form
                       :math:`z = \Sigma (x_i \cdot \z_i)`
        'power         Applies an exponent to the property as:
                       :math:`\z^{power} = \Sigma (x_i \cdot \z_i^{power})`
        ============== ========================================================

    power : scalar
        If ``mode='power'`` this indicates the value of the exponent,
        otherwise this is ignored.
    """
    xs = phase['pore.mole_fraction']
    ys = phase.get_comp_vals(prop)
    z = 0.0
    if mode == 'logarithmic':
        for i in xs.keys():
            z += xs[i]*np.log(ys[i])
        z = np.exp(z)
    elif mode in ['linear', 'simple']:
        for i in xs.keys():
            z += xs[i]*ys[i]
    elif mode == 'power':
        for i in xs.keys():
            z += xs[i]*ys[i]**power
        z = z**(1/power)
    return z


@_phasedocs
def mole_to_mass_fraction(
    phase,
    MWs='param.molecular_weight'
):
    r"""
    Computes the mass fraction in each pore

    Parameters
    ----------
    %(phase)s
    %(MWs)s

    Returns
    -------
    ws : ndarray
        An ndarray containing the mass fraction in each pore computed from
        the mole fractions of each component and their molecular weights.
    """
    xs = phase['pore.mole_fraction']
    MW = phase.get_comp_vals(MWs)
    num = [xs[k]*MW[k] for k in xs.keys()]
    denom = np.sum(num, axis=0)
    ws = num/denom
    comps = list(phase.components.keys())
    d = {comps[i]: ws[i, :] for i in range(len(comps))}
    return d


@_phasedocs
def mole_summation(phase):
    r"""
    Computes total mole fraction in each pore given component values

    Parameters
    ----------
    %(phase)s

    Returns
    -------
    vals : ND-array
        An ND-array containing the total mole fraction.  Note that this is not
        guaranteed to sum to 1.0.
    """
    xs = [phase['pore.mole_fraction.' + c.name] for c in phase.components.values()]
    if len(xs) > 0:
        xs = np.sum(xs, axis=0)
    else:
        xs = np.nan
    return xs


@_phasedocs
def from_component(phase, prop, compname):
    r"""
    Fetches the given values from the specified object

    Parameters
    ----------
    %(phase)s
    prop : str
        The name of the array to retreive
    compname : str
        The name of the object possessing the desired data

    Returns
    -------
    vals : ND-array
        An ND-array containing the request data
    """
    comp = phase.project[compname]
    vals = comp[prop]
    return vals



r"""
Geometry
--------

This submodule contains pore-scale models that calculate geometrical
properties. These models are to be added to a Geometry object.

"""
from ._geodocs import *
from . import pore_size
from . import pore_seed
from . import pore_volume
from . import pore_surface_area
from . import pore_cross_sectional_area
from . import throat_cross_sectional_area
from . import throat_seed
from . import throat_size
from . import throat_length
from . import throat_perimeter
from . import throat_surface_area
from . import throat_volume
from . import throat_capillary_shape_factor
from . import throat_centroid
from . import throat_endpoints
from . import throat_vector
from . import hydraulic_size_factors
from . import diffusive_size_factors
from . import conduit_lengths



import logging
import numpy as np
from pathlib import Path


__all__ = [
    '_parse_filename',
    '_parse_args',
]


logger = logging.getLogger(__name__)


def _parse_filename(filename, ext=""):
    p = Path(filename)
    p = p.resolve()
    if ext == "":
        ext = p.suffix
    # If extension not part of filename
    ext = "." + (ext.strip("."))
    if p.suffix != ext:
        p = p.with_suffix(ext)
    return p


def _parse_args(network, phases):
    try:
        project = network.project
        network = [network]
    except AttributeError:
        project = network[0].project
    # Ensure phases is a list, even if empty
    if not isinstance(phases, list):
        phases = [phases]
    return (project, network, phases)



from openpnm.network import Network
from openpnm.utils import Docorator
from openpnm._skgraph.generators import delaunay


__all__ = ['Delaunay']
docstr = Docorator()


@docstr.dedent
class Delaunay(Network):
    r"""
    Random network formed by Delaunay tessellation of arbitrary base points

    Parameters
    ----------
    points : array_like or int
        Can either be an N-by-D array of point coordinates which will be used,
        or a scalar value indicating the number of points to generate. If 3D points
        are supplied, and a 2D shape is specified, then the z-coordinate is ignored.
    shape : array_like
        The size and shape of the domain:

        ========== ============================================================
        shape      result
        ========== ============================================================
        [x, y, z]  A 3D cubic domain of dimension x, y and z
        [x, y, 0]  A 2D square domain of size x by y
        ========== ============================================================

    reflect : bool, optional
        If ``True`` then the base points will be reflected across
        all the faces of the domain prior to performing the tessellation. This
        feature is best combined with ``trim=True`` to prevent unreasonably long
        connections between points on the surfaces.
    f : float
        The fraction of points which should be reflected.  The default is 1 which
        reflects all the points in the domain, but this can lead to a lot of
        unnecessary points, so setting to 0.1 or 0.2 helps speed, but risks that
        the tessellation may not have smooth faces if not enough points are
        reflected.
    trim : bool, optional
        If ``True`` (default) then all vertices laying outside the domain will
        be removed. This is only useful if ``reflect=True``.

    %(Network.parameters)s

    Notes
    -----
    It is also possible to generate circular ``[r, 0]``, cylindrical ``[r, z]``, and
    spherical domains ``[r]``, but this feature does not quite work as desired.
    It does not produce a truly clean outer surface since the tessellation are
    conducted in cartesian coordinates, so the circular and spherical surfaces
    have artifacts. Scipy recently added the ability to do tessellations on
    spherical surfaces, for geological applications, but this is not flexible
    enough, yet.

    """

    def __init__(self, shape, points, reflect=True, trim=True, f=1, **kwargs):
        super().__init__(**kwargs)
        net, tri = delaunay(
            points=points,
            shape=shape,
            reflect=reflect,
            f=f,
            trim=trim,
            node_prefix='pore',
            edge_prefix='throat',
        )
        self.update(net)
        self._post_init()
        self.tri = tri



import openpnm.models as mods


air = {
    'pore.density': {
        'model': mods.phase.density.ideal_gas,
    },
    'pore.molar_density': {
        'model': mods.phase.density.mass_to_molar,
    },
    'pore.diffusivity': {
        'model': mods.phase.diffusivity.gas_mixture_fesg,
        'MWs': [31.9988, 28.0134],
        'Vdms': [16.6, 17.9],
    },
    'pore.thermal_conductivity': {
        'model': mods.misc.polynomial,
        'prop': 'pore.temperature',
        'a': [0.00422791, 0.0000789606, -1.56383E-08],
    },
    'pore.viscosity': {
        'model': mods.misc.polynomial,
        'prop': 'pore.temperature',
        'a': [0.00000182082, 6.51815E-08, -3.48553E-11, 1.11409E-14],
    },
}



import os
import logging
import numpy as np
from openpnm.topotools import trim, extend
from openpnm.network import Network
from pathlib import Path
from pandas import read_table, DataFrame
from tqdm.auto import tqdm


logger = logging.getLogger(__name__)


def network_from_statoil(path, prefix):
    r"""
    Load data from the \'dat\' files located in specified folder.

    Parameters
    ----------
    path : str
        The full path to the folder containing the set of \'dat\' files.
    prefix : str
        The file name prefix on each file. The data files are stored
        as \<prefix\>_node1.dat.
    network : Network
        If given then the data will be loaded on it and returned.  If not
        given, a Network will be created and returned.

    Returns
    -------
    Project
        An OpenPNM Project containing a Network holding all the data

    Notes
    -----
    The StatOil format is used by the Maximal Ball network extraction code of
    the Imperial College London group

    This class can be used to load and work with those networks. Numerous
    datasets are available for download from the group's
    `website <http://tinyurl.com/zurko4q>`_.

    The 'Statoil' format consists of 4 different files in a single
    folder. The data is stored in columns with each corresponding to a
    specific property. Headers are not provided in the files, so one must
    refer to various theses and documents to interpret their meaning.

    """
    net = {}

    # Parse the link1 file
    path = Path(path)
    filename = Path(path.resolve(), prefix+'_link1.dat')
    with open(filename, mode='r') as f:
        link1 = read_table(filepath_or_buffer=f,
                           header=None,
                           skiprows=1,
                           sep=' ',
                           skipinitialspace=True,
                           index_col=0)
    link1.columns = ['throat.pore1', 'throat.pore2', 'throat.radius',
                     'throat.shape_factor', 'throat.total_length']
    # Add link1 props to net
    net['throat.conns'] = np.vstack((link1['throat.pore1']-1,
                                     link1['throat.pore2']-1)).T
    net['throat.conns'] = np.sort(net['throat.conns'], axis=1)
    net['throat.radius'] = np.array(link1['throat.radius'])
    net['throat.shape_factor'] = np.array(link1['throat.shape_factor'])
    net['throat.total_length'] = np.array(link1['throat.total_length'])

    filename = Path(path.resolve(), prefix+'_link2.dat')
    with open(filename, mode='r') as f:
        link2 = read_table(filepath_or_buffer=f,
                           header=None,
                           sep=' ',
                           skipinitialspace=True,
                           index_col=0)
    link2.columns = ['throat.pore1', 'throat.pore2',
                     'throat.pore1_length', 'throat.pore2_length',
                     'throat.length', 'throat.volume',
                     'throat.clay_volume']
    # Add link2 props to net
    cl_t = np.array(link2['throat.length'])
    net['throat.length'] = cl_t
    net['throat.conduit_lengths.throat'] = cl_t
    net['throat.volume'] = np.array(link2['throat.volume'])
    cl_p1 = np.array(link2['throat.pore1_length'])
    net['throat.conduit_lengths.pore1'] = cl_p1
    cl_p2 = np.array(link2['throat.pore2_length'])
    net['throat.conduit_lengths.pore2'] = cl_p2
    net['throat.clay_volume'] = np.array(link2['throat.clay_volume'])
    # ---------------------------------------------------------------------
    # Parse the node1 file
    filename = Path(path.resolve(), prefix+'_node1.dat')
    with open(filename, mode='r') as f:
        row_0 = f.readline().split()
        num_lines = int(row_0[0])
        array = np.ndarray([num_lines, 6])
        for i in range(num_lines):
            row = f.readline()\
                   .replace('\t', ' ').replace('\n', ' ').split()
            array[i, :] = row[0:6]
    node1 = DataFrame(array[:, [1, 2, 3, 4]])
    node1.columns = ['pore.x_coord', 'pore.y_coord', 'pore.z_coord',
                     'pore.coordination_number']
    # Add node1 props to net
    net['pore.coords'] = np.vstack((node1['pore.x_coord'],
                                    node1['pore.y_coord'],
                                    node1['pore.z_coord'])).T
    # ---------------------------------------------------------------------
    # Parse the node2 file
    filename = Path(path.resolve(), prefix+'_node2.dat')
    with open(filename, mode='r') as f:
        node2 = read_table(filepath_or_buffer=f,
                           header=None,
                           sep=' ',
                           skipinitialspace=True,
                           index_col=0)
    node2.columns = ['pore.volume', 'pore.radius', 'pore.shape_factor',
                     'pore.clay_volume']
    # Add node2 props to net
    net['pore.volume'] = np.array(node2['pore.volume'])
    net['pore.radius'] = np.array(node2['pore.radius'])
    net['pore.shape_factor'] = np.array(node2['pore.shape_factor'])
    net['pore.clay_volume'] = np.array(node2['pore.clay_volume'])
    net['throat.cross_sectional_area'] = ((net['throat.radius']**2)
                                          / (4.0*net['throat.shape_factor']))
    net['pore.area'] = ((net['pore.radius']**2)
                        / (4.0*net['pore.shape_factor']))

    network = Network()
    network.update(net)

    # Use OpenPNM Tools to clean up network
    # Trim throats connected to 'inlet' or 'outlet' reservoirs
    trim1 = np.where(np.any(net['throat.conns'] == -1, axis=1))[0]
    # Apply 'outlet' label to these pores
    outlets = network['throat.conns'][trim1, 1]
    network['pore.outlets'] = False
    network['pore.outlets'][outlets] = True
    trim2 = np.where(np.any(net['throat.conns'] == -2, axis=1))[0]
    # Apply 'inlet' label to these pores
    inlets = network['throat.conns'][trim2, 1]
    network['pore.inlets'] = False
    network['pore.inlets'][inlets] = True
    # Now trim the throats
    to_trim = np.hstack([trim1, trim2])
    trim(network=network, throats=to_trim)

    return network



import logging
import pickle
import re
import sys
from datetime import datetime
from uuid import uuid4

from openpnm.utils import SettingsAttr

logger = logging.getLogger("openpnm")

__all__ = ['Workspace']


class WorkspaceSettings(SettingsAttr):
    r"""

    Parameters
    ----------
    default_solver : str
        The solver to use by default, if user does not specify one explicitly.
        The default values is PardisoSpsolve, but a good option is ScipySpsolve
        if the Pardiso is causing problems.
    loglevel : int
        Sets the threshold for the severity of logger message which appear.
        Ranges are as follows:

        ======= ==============================================================
        Level   Description
        ======= ==============================================================
        10      DEBUG: Detailed information, typically of interest only when
                diagnosing problems.
        20      INFO: Confirmation that things are working as expected.
        30      WARNING: An indication that something unexpected happened, or
                indicative of some problem in the near future (e.g. ‘disk
                space low’). The software is still working as expected.
        40      ERROR: Due to a more serious problem, the software has not
                been able to perform some function.
        50      CRITICAL: A serious error, indicating that the program itself
                may be unable to continue running.
        ======= ==============================================================
    """
    # Pardiso requires MKL, which is not available on new Apple chips
    if sys.platform == 'darwin':
        default_solver = 'ScipySpsolve'
    else:
        try:
            import pypardiso
            default_solver = 'PardisoSpsolve'
        except ImportError:
            default_solver = 'ScipySpsolve'
            msg = (
                'PARDISO solver not installed on this platform. '
                'Simulations will be slow.'
            )
            logger.warning(msg)

    @property
    def loglevel(self):
        return self._loglevel

    @loglevel.setter
    def loglevel(self, value):
        if isinstance(value, str):
            options = {
                "TRACE": 5,
                "DEBUG": 10,
                "INFO": 20,
                "SUCESS": 25,
                "WARNING": 30,
                "ERROR": 40,
                "CRITICAL": 50
            }
            value = options[value]
        self._loglevel = value
        logger.setLevel(value)


class Workspace(dict):
    r"""
    The Workspace object provides the highest level of adminstrative
    control over active OpenPNM sessions.

    It is a ``dictionary`` that stores a list of all open Projects by name.

    This class is a
    `singleton <https://en.wikipedia.org/wiki/Singleton_pattern>`_ so that
    whenever and wherever a Workspace is instantiated, the same instance is
    obtained. This allows it to maintain a definitive record of all open
    Projects.

    See Also
    --------
    Project

    Notes
    -----
    The Workspace object contains a variety of functions that one might
    expect from the 'file-menu' in a typical GUI.

    """

    __instance__ = None

    # This __new__ method makes the Workspace a Singleton
    def __new__(cls, *args, **kwargs):
        if Workspace.__instance__ is None:
            Workspace.__instance__ = dict.__new__(cls)
        return Workspace.__instance__

    def __init__(self):
        super().__init__()
        self.settings = WorkspaceSettings()
        self.settings.loglevel = 30

    def copy(self):
        """Brief explanation of 'copy'"""
        raise Exception('Cannot copy Workspace, only one can exist at a time')

    def _create_console_handles(self, project):
        r"""
        Adds all objects in the given project to the console as variables
        with handle names taken from each object's name.
        """
        import __main__
        for item in project:
            __main__.__dict__[item.name] = item

    def _validate_name(self, name=None):
        r"""
        Generates a valid name for projects
        """
        if name in [None, '']:
            name = 'proj_01'  # Give basic name, then let rest of func fix it
        if name in self.keys():  # If proposed name is taken, increment it
            if not re.search(r'_\d+$', name):  # If name does not end with _##
                name = name + '_01'
            prefix, count = name.rsplit('_', 1)
            n = [0]
            for item in list(self.keys()):
                if item.startswith(prefix+'_'):
                    n.append(int(item.split(prefix+'_')[1]))
            name = prefix+'_'+str(max(n)+1).zfill(2)
        return name

    def save_workspace(self, filename=None):
        r"""
        Saves all projects in the current workspace as a single file

        Parameters
        ----------
        filename : str
            The filename to use when saving. If not provided, the present
            date and time are used.

        Notes
        -----
        The file is actually zip archive containing ``pnm`` files, one for
        each project in the workspace. This archive can be extracted and each
        ``pnm`` file can be loaded manually using ``load_project`` or the
        ``openpnm.io.PNM`` class.

        """
        if filename is None:
            dt = datetime.now()
            filename = dt.strftime("%Y_%m_%d_%H_%M_%S")
        from zipfile import ZipFile
        with ZipFile(filename + '.wrk', 'w') as z:
            for prj in self.values():
                prj.save_project()
                z.write(prj.name + '.pnm')

    def load_workspace(self, filename):
        r"""
        Loads project(s) from a saved workspace into current workspace

        Parameters
        ----------
        filename : str or Path
            The filename containing the saved workspace

        """
        from zipfile import ZipFile
        with ZipFile(filename, 'r') as z:
            logger.info('Loading projects contained in ' + filename)
            files = z.filelist
            for f in files:
                self.load_project(f.orig_filename)

    def save_project(self, project, filename=None):
        r"""
        Saves given Project to a ``pnm`` file

        This will include all of associated objects, including algorithms.

        Parameters
        ----------
        project : Project
            The project to save.
        filename : str, optional
            If no filename is given, the given project name is used. See Notes
            for more information.

        See Also
        --------
        save_workspace

        Notes
        -----
        The filename can be a string such as 'saved_file.pnm'.  The string can
        include absolute path such as 'C:\networks\saved_file.pnm', or can
        be a relative path such as '..\..\saved_file.pnm', which will look
        2 directories above the current working directory.  Can also be a
        path object object such as that produced by ``pathlib`` or
        ``os.path`` in the Python standard library.

        """
        if filename is None:
            dt = datetime.now()
            filename = dt.strftime("%Y_%m_%d_%H_%M_%S")
        with open(filename.split('.')[-1]+'.pnm', 'wb') as f:
            pickle.dump(project, f)

    def load_project(self, filename):
        r"""
        Loads a Project from the specified 'pnm' file

        The loaded project is added to the Workspace. This will *not* delete
        any existing Projects in the Workspace and will rename any Projects
        being loaded if necessary.

        Parameters
        ----------
        filename : str or Path
            The name of the file to open. See Notes for more information.

        See Also
        --------
        load_workspace

        Notes
        -----
        The filename can be a string such as 'saved_file.pnm'. The string can
        include absolute path such as 'C:\networks\saved_file.pnm', or can
        be a relative path such as '..\..\saved_file.pnm', which will look
        2 directories above the current working directory.  Can also be a
        path object object such as that produced by ``pathlib`` or
        ``os.path`` in the Python standard library.

        """
        with open(filename, 'rb') as f:
            proj = pickle.load(f)
            proj.settings.uuid = str(uuid4())
            proj.name = self._validate_name(proj.name)
            self[proj.name] = proj
        return proj

    def close_project(self, project):
        r"""
        Removes the specified Project from the Workspace

        This does not save the project, so any changes will be lost.

        Parameters
        ----------
        project : Project
            The Project object to be copied

        """
        _ = self.pop(project.name, None)

    def copy_project(self, project, name=None):
        r"""
        Makes a copy of an existing Project

        Parameters
        ----------
        project : Project
            The Project object to be copied
        name : str, optional
            A name for the new copy of the project. If not supplied, then
            one will be generated (e.g. 'proj_02')

        Returns
        -------
        proj : list
            A handle to the new Project

        """
        proj = project.copy(name)
        return proj

    def new_project(self, name=None):
        r"""
        Creates a new empty Project object

        Parameters
        ----------
        name : str, optional
            The unique name to give to the project. If none is given, one
            will be automatically generated (e.g. 'proj_01`)

        Returns
        -------
        proj : list
            An empty Project object, suitable for passing into a Network
            generator

        """
        from openpnm.utils import Project
        proj = Project(name=name)
        return proj

    def __str__(self):  # pragma: no cover
        hr = '―'*78
        s = hr + '\n'
        for item in self.values():
            s += item.name + '\n'
            s += item.__str__() + '\n'
        return s
